from dataclasses import dataclass, field
from typing import List, Optional, Dict, Any
from enum import Enum, auto
from acceldata_sdk.constants import FailureStrategy
from acceldata_sdk.models.ruleExecutionResult import RuleExecutionSummary, RuleExecutionResult
from acceldata_sdk.models.common_types import PolicyExecutionRequest
from acceldata_sdk.models.threshold import RuleLevelThresholdConfiguration


class Rule:
    def __init__(self, id, name, description, createdAt, updatedAt, backingAssets, thresholdLevel, archivalReason=None,
                 archived=False, enabled=False, schedule=None, notificationChannels=None, *args, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.backingAssets = backingAssets
        self.thresholdLevel = thresholdLevel
        self.archivalReason = archivalReason
        self.archived = archived
        self.enabled = enabled
        self.schedule = schedule
        self.notificationChannels = notificationChannels

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"Rule({self.__dict__})"


class RuleCancelResponse:

    def __init__(self, message, status, *args, **kwargs):
        self.status = status
        self.message = message

    def __repr__(self):
        return f"Response({self.__dict__})"


class RuleExecution:

    def __init__(self, ruleId, ruleName, thresholdLevel, startedAt, resultStatus, executionStatus, executionMode
                 , ruleType=None, id=None, lastMarker=None, executionError=None, rightLastMarker=None,
                 leftLastMarker=None, finishedAt=None, ruleVersion=None, *args, **kwargs):
        self.ruleId = ruleId
        self.ruleName = ruleName
        self.ruleVersion = ruleVersion
        self.thresholdLevel = thresholdLevel
        self.startedAt = startedAt
        self.resultStatus = resultStatus
        self.executionStatus = executionStatus
        self.executionMode = executionMode
        self.ruleType = ruleType
        self.executionError = executionError
        self.rightLastMarker = rightLastMarker
        self.lastMarker = lastMarker
        self.leftLastMarker = leftLastMarker
        self.id = id
        self.finishedAt = finishedAt

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"RuleExecution({self.__dict__})"

class ExecutionResult:

    def __init__(self, execution, items, meta=None, result=None, *args, **kwargs):
        if isinstance(execution, dict):
            self.execution = RuleExecution(**execution)
        else:
            self.execution = execution
        if isinstance(result, dict):
            self.result = RuleExecutionResult(**result)
        else:
            self.result = result
        self.items = items
        self.meta = meta

    def __repr__(self):
        return f"ExecutionResult({self.__dict__})"


@dataclass
class ShortSegment:
    id = None
    name = None

    def __init__(self, id, name, *args, **kwargs):
        self.id = id
        self.name = name

    def __repr__(self):
        return f"ShortSegment({self.__dict__})"


@dataclass
class Label:
    def __init__(self, key, value, *args, **kwargs):
        self.key = key
        self.value = value

    def __repr__(self):
        return f"Label({self.__dict__})"


@dataclass
class BackingAsset:
    id = None
    ruleId = None
    tableAssetId = None
    customQuery = None
    customTableAssetIds = None
    tableAlias = None
    marker = None

    def __init__(self, id=None, ruleId=None, tableAssetId=None, customQuery=None, customTableAssetIds=None,
                 tableAlias=None, marker=None, *args, **kwargs):
        self.id = id
        self.ruleId = ruleId
        self.tableAssetId = tableAssetId
        self.customQuery = customQuery
        self.customTableAssetIds = customTableAssetIds
        self.tableAlias = tableAlias
        self.marker = marker

    def __repr__(self):
        return f"BackingAsset({self.__dict__})"


@dataclass
class RuleTag:
    id = None,
    tagId = None
    ruleId = None
    name = None

    def __init__(self, id, tagId, ruleId, name, *args, **kwargs):
        self.id = id
        self.tagId = tagId
        self.ruleId = ruleId
        self.name = name

    def __repr__(self):
        return f"RuleTag({self.__dict__})"


@dataclass
class PolicyGroup:
    id = None
    name = None
    description = None
    createdAt = None
    updatedAt = None
    tenantId = None

    def __init__(self, id, name, description, createdAt, updatedAt, tenantId, *args, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.createdAt = createdAt
        self.updatedAt = updatedAt
        self.tenantId = tenantId

    def __repr__(self):
        return f"PolicyGroup({self.__dict__})"

class RuleResource:
    def __init__(self, rule, details, client=None, *args, **kwargs):
        self.rule = rule
        self.details = details
        self.client = client

    def execute(self, sync=True, incremental=False, failure_strategy: FailureStrategy = FailureStrategy.DoNotFail,
                policy_execution_request: PolicyExecutionRequest = None) -> RuleExecution:
        pass

    def get_executions(self, page=0, size=25, sortBy='finishedAt:DESC') -> List[RuleExecutionSummary]:
        pass



import datetime


class DateTimeFormatter:

    @staticmethod
    def format_datetime(date_time):
        if date_time is None:
            return None
        if isinstance(date_time, datetime.datetime):
            return date_time.isoformat()
        else:
            raise ValueError("Input is not a valid datetime object")



import logging
import pytest

import acceldata_sdk.constants as const
from acceldata_sdk.constants import FailureStrategy
from acceldata_sdk.models.ruleExecutionResult import PolicyFilter, RuleType

from . import test_constants as test_const
from ..commons.retry import retry_operation

logger = logging.getLogger(__name__)

FRESHNESS_POLICY_NAME = test_const.FRESHNESS_POLICY_NAME


# =====================================================
# Get freshness policy by name
# =====================================================
@pytest.mark.integration
def test_get_freshness_policy_by_name(adoc_client):
    def operation():
        return adoc_client.get_policy(
            const.PolicyType.DATA_CADENCE,
            FRESHNESS_POLICY_NAME
        )

    policy = retry_operation(
        operation,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    logger.info("Fetched freshness policy by name: %s", policy)
    assert policy is not None


# =====================================================
# List freshness policies by filter
# =====================================================
@pytest.mark.integration
def test_get_freshness_policy_by_filter(adoc_client):
    policy_filter = PolicyFilter(
        policyType=RuleType.DATA_CADENCE,
        enable=True
    )

    def operation():
        return adoc_client.list_all_policies(filter=policy_filter)

    policies = retry_operation(
        operation,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    logger.info("Fetched freshness policies: %s", policies)
    assert policies is not None


# =====================================================
# Execute freshness policy
# =====================================================
@pytest.fixture(scope="module")
def freshness_execution_id(adoc_client):
    def get_policy_op():
        return adoc_client.get_policy(identifier=FRESHNESS_POLICY_NAME)

    policy = retry_operation(
        get_policy_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    def execute_op():
        return adoc_client.execute_freshness_rule(
            rule_id=policy.id
        )

    execution = retry_operation(
        execute_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    logger.info("Freshness policy execution started: %s", execution)
    return execution.id





# =====================================================
# Get freshness rule result
# =====================================================
@pytest.mark.integration
def test_freshness_policy_rule_result(
        adoc_client,
        freshness_execution_id
):
    logger.info("Freshness execution id: %s",freshness_execution_id)
    def operation():
        return adoc_client.get_freshness_rule_result(
            freshness_execution_id
        )

    result = retry_operation(
        operation,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    logger.info("Freshness rule result: %s", result)
    assert result is not None


# =====================================================
# Execute policy sync
# =====================================================
@pytest.mark.integration
def test_execute_policy_sync(adoc_client):
    def get_policy_op():
        return adoc_client.get_policy(identifier=FRESHNESS_POLICY_NAME)

    policy = retry_operation(
        get_policy_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    def execute_op():
        return adoc_client.execute_policy(
            const.PolicyType.DATA_CADENCE,
            policy.id,
            sync=True,
            failure_strategy=FailureStrategy.DoNotFail
        )

    executor = retry_operation(
        execute_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    def result_op():
        return executor.get_result(
            failure_strategy=FailureStrategy.DoNotFail
        )

    result = retry_operation(
        result_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    logger.info("Sync execution result: %s", result)
    logger.info("Execution status: %s", executor.get_status())


# =====================================================
# Execute async and cancel
# =====================================================
@pytest.mark.integration
def test_execute_policy_async_and_cancel(adoc_client):
    def get_policy_op():
        return adoc_client.get_policy(identifier=FRESHNESS_POLICY_NAME)

    policy = retry_operation(
        get_policy_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    def execute_op():
        return adoc_client.execute_policy(
            const.PolicyType.DATA_CADENCE,
            policy.id,
            sync=False,
            failure_strategy=FailureStrategy.DoNotFail
        )

    async_executor = retry_operation(
        execute_op,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL
    )

    # Important safety check
    if not async_executor or not getattr(async_executor, "id", None):
        logger.warning(
            "Skipping cancel — async execution did not start"
        )
        return

    async_executor.cancel()
    logger.info("Async execution cancelled")

    logger.info("Async execution cancelled", )



import logging
import os
import sys

import pytest


def pytest_sessionstart(session):
    """
    Matches tests/conftest.py pytest_sessionstart().
    Must be called BEFORE importing acceldata_sdk.
    """

    root = logging.getLogger()

    # remove any existing handlers
    for h in list(root.handlers):
        root.removeHandler(h)

    root.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        "%(asctime)s | %(levelname)s | %(name)s | %(message)s"
    )

    handler.setFormatter(formatter)
    root.addHandler(handler)


@pytest.fixture(scope="session")
def adoc_client():
    from acceldata_sdk.torch_client import TorchClient

    return TorchClient(
        url=os.environ["TORCH_CATALOG_URL"],
        access_key=os.environ["TORCH_ACCESS_KEY"],
        secret_key=os.environ["TORCH_SECRET_KEY"],
        torch_connection_timeout_ms=int(os.getenv("TORCH_CONNECTION_TIMEOUT_MS", "40000")),
        torch_read_timeout_ms=int(os.getenv("TORCH_READ_TIMEOUT_MS", "40000")),
        do_version_check=os.getenv("DO_VERSION_CHECK", "false").lower() == "true",
    )



import unittest
from datetime import datetime, timedelta
from acceldata_sdk.time_range_utils import TimeRangeCalculator
from acceldata_sdk.models.ruleExecutionResult import ExecutionPeriod
from acceldata_sdk.models.ruleExecutionResult import PolicyFilter
from dateutil.relativedelta import relativedelta

class TestCalculateTimeRange(unittest.TestCase):
    def test_last15minutes(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last15minutes)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(minutes=15)
        print("ExecutionPeriod.Last15minutes")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last30minutes(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last30minutes)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(minutes=30)
        print("ExecutionPeriod.Last30minutes")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last_1_hour(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last1hour)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(hours=1)
        print("ExecutionPeriod.Last1hour")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last_3_hour(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last3hours)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(hours=3)
        print("ExecutionPeriod.Last3hours")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last_6_hour(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last6hours)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(hours=6)
        print("ExecutionPeriod.Last6hours")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last_12_hour(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last12hours)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(hours=12)
        print("ExecutionPeriod.Last12hours")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_last_24_hour(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last24hours)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_finished_before = datetime.now()
        expected_started_after = expected_finished_before - timedelta(hours=24)
        print("ExecutionPeriod.Last24hours")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_execution_period_today(self):
        filter = PolicyFilter(period=ExecutionPeriod.Today)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)

        now = datetime.now()
        expected_started_after = datetime.combine(now.date(), datetime.min.time())
        expected_finished_before = now

        print("ExecutionPeriod.Today")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)

        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_execution_period_yesterday(self):
        filter = PolicyFilter(period=ExecutionPeriod.Yesterday)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)

        now = datetime.now()
        expected_finished_before = datetime.combine(now.date(), datetime.min.time())
        expected_started_after = datetime.combine((now - timedelta(days=1)).date(), datetime.min.time())
        print("ExecutionPeriod.Yesterday")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)


    def test_execution_period_last_7_days(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last7days)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)
        expected_started_after = datetime.combine((datetime.now() - timedelta(days=7)).date(), datetime.min.time())
        expected_finished_before = datetime.combine(datetime.now().date(), datetime.min.time())
        print("ExecutionPeriod.Last7days")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)


    def test_execution_period_this_month(self):
        filter = PolicyFilter(period=ExecutionPeriod.Thismonth)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)

        now = datetime.now()
        expected_finished_before = now
        expected_started_after = datetime(datetime.now().year, datetime.now().month, 1)
        print("ExecutionPeriod.Thismonth")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")

        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)
        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)


    def test_execution_period_last3month(self):
        filter = PolicyFilter(period=ExecutionPeriod.Last3month)
        finished_before, started_after = TimeRangeCalculator.calculate_time_range(filter)

        now = datetime.now()
        expected_started_after = (now - relativedelta(months=3)).replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        expected_finished_before = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        print("ExecutionPeriod.Last3month")
        print("started_after: "+str(started_after))
        print("finished_before: "+str(finished_before))
        print("===========================================")

        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)

        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

    def test_calculate_days_before_time(self):
        days = 7  # Adjust the number of days as needed
        started_after, finished_before = TimeRangeCalculator.calculate_days_before_time(days)

        now = datetime.now()
        expected_started_after = datetime.combine((now - timedelta(days=days)).date(), datetime.min.time())
        expected_finished_before = datetime.combine(now.date(), datetime.min.time())

        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)


    def test_calculate_hours_before_time(self):
        hours = 6  # Adjust the number of hours as needed
        started_after, finished_before = TimeRangeCalculator.calculate_hours_before_time(hours)

        now = datetime.now()
        expected_started_after = now - timedelta(hours=hours)
        expected_finished_before = now

        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)

        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)

    def test_calculate_minutes_before_time(self):
        minutes = 30  # Adjust the number of minutes as needed
        started_after, finished_before = TimeRangeCalculator.calculate_minutes_before_time(minutes)

        now = datetime.now()
        expected_started_after = now - timedelta(minutes=minutes)
        expected_finished_before = now

        self.assertEqual(started_after.year, expected_started_after.year)
        self.assertEqual(started_after.month, expected_started_after.month)
        self.assertEqual(started_after.day, expected_started_after.day)
        self.assertEqual(started_after.hour, expected_started_after.hour)
        self.assertEqual(started_after.minute, expected_started_after.minute)
        self.assertEqual(finished_before.year, expected_finished_before.year)
        self.assertEqual(finished_before.month, expected_finished_before.month)
        self.assertEqual(finished_before.day, expected_finished_before.day)
        self.assertEqual(finished_before.hour, expected_finished_before.hour)
        self.assertEqual(finished_before.minute, expected_finished_before.minute)

    def test_calculate_months_before_time(self):
        months_before = 3  # Adjust the number of months as needed
        started_after, finished_before = TimeRangeCalculator.calculate_months_before_time(months_before)

        now = datetime.now()

        # Calculate the expected first day of 'months_before' months ago
        expected_first_day_of_months_before = (now - relativedelta(months=months_before)).replace(day=1, hour=0, minute=0, second=0, microsecond=0)

        # Calculate the expected last day of the previous month
        expected_last_day_of_prev_month = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)

        self.assertEqual(started_after, expected_first_day_of_months_before)
        self.assertEqual(finished_before, expected_last_day_of_prev_month)


if __name__ == '__main__':
    unittest.main()



from acceldata_sdk.events.span_event import SpanEvent
from datetime import datetime


class LogEvent(SpanEvent):
    """
        Description:
            Class to send log event to torch.
    """

    def __init__(self, log_data: str, context_data=None, created_at: datetime = None):
        """
        :param log_data: log data of the log event
        :param context_data: context data of event
        :param event_uid: event uid
        :param created_at: An optional datetime representing the explicit time for sending the log event.
         If not set, the current time will be used to send the log event
        """
        self.log_data = log_data
        self.context_data = context_data
        self.event_uid = 'LOG'
        self.created_at = created_at



import logging
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any


from acceldata_sdk.constants import FailureStrategy, PolicyType
from acceldata_sdk.models.common_types import PolicyExecutionRequest
from acceldata_sdk.models.threshold import parse_rule_threshold_configuration
from acceldata_sdk.models.rule import (
    ShortSegment,
    BackingAsset,
    RuleTag,
    PolicyGroup,
    RuleResource,
    RuleExecution,
)
from acceldata_sdk.models.ruleExecutionResult import (
    RuleExecutionResult,
    RuleItemResult,
    RuleExecutionSummary,
    RuleResult,
    ExecutorConfig,
    SparkResourceConfig,
    RuleExecutionResult
)

logger = logging.getLogger("freshnessrule")
logger.setLevel(logging.INFO)


# --------------------------------------------------------------------
# Item (aligned to details.items[])
# --------------------------------------------------------------------
@dataclass
class Item:
    id: Optional[int] = None
    ruleItemId: Optional[int] = None
    threshold: Optional[float] = None
    resultPercent: Optional[float] = None
    success: Optional[bool] = None
    error: Optional[str] = None
    weightage: Optional[float] = None
    isWarning: Optional[bool] = None
    dimension: Optional[str] = None
    includeInQualityScore: Optional[bool] = None
    anomalyDetected: Optional[bool] = None
    anomalyDetails: Optional[List[Dict[str, Any]]] = field(default_factory=list)
    thresholdBreached: Optional[bool] = None
    thresholdBreachDetails: Optional[Dict[str, Any]] = field(default_factory=dict)

    def __init__(
            self,
            id=None,
            ruleItemId=None,
            threshold=None,
            resultPercent=None,
            success=None,
            error=None,
            weightage=None,
            isWarning=None,
            dimension=None,
            includeInQualityScore=None,
            anomalyDetected=None,
            anomalyDetails=None,
            thresholdBreached=None,
            thresholdBreachDetails=None,
            *args,
            **kwargs,
    ):
        self.id = id
        self.ruleItemId = ruleItemId
        self.threshold = threshold
        self.resultPercent = resultPercent
        self.success = success
        self.error = error
        self.weightage = weightage
        self.isWarning = isWarning
        self.dimension = dimension
        self.includeInQualityScore = includeInQualityScore
        self.anomalyDetected = anomalyDetected
        self.anomalyDetails = anomalyDetails or []
        self.thresholdBreached = thresholdBreached
        self.thresholdBreachDetails = thresholdBreachDetails or {}

    def __repr__(self):
        return f"Item({self.__dict__})"


class FreshnessRuleExecutionSummary(RuleExecutionSummary):
    def __init__(self, ruleId, executionMode, executionStatus, resultStatus, startedAt, executionType,
                 isProtectedResource, thresholdLevel, ruleVersion, id=None, ruleName=None, ruleType=None,
                 lastMarker=None, leftLastMarker=None, rightLastMarker=None, executionError=None, finishedAt=None,
                 resetPoint=None, persistencePath=None, resultPersistencePath=None, executorConfig=None,
                 markerConfig=None, leftMarkerConfig=None, rightMarkerConfig=None, dataPersistenceEnabled=None,
                 sparkResourceConfig=None, *args, **kwargs):
        super().__init__(ruleId, executionMode, executionStatus, resultStatus, startedAt, executionType, thresholdLevel,
                         ruleVersion, id, ruleName, ruleType, lastMarker, leftLastMarker, rightLastMarker,
                         executionError, finishedAt, resetPoint, persistencePath, resultPersistencePath, executorConfig,
                         sparkResourceConfig, markerConfig, leftMarkerConfig, rightMarkerConfig, dataPersistenceEnabled,
                         isProtectedResource)

    def __repr__(self):
        return f"FreshnessRuleExecutionSummary({self.__dict__})"


class FreshnessRuleExecutionResult(RuleExecutionResult):
    def __init__(
            self,
            status,
            description=None,
            successCount=None,
            failureCount=None,
            qualityScore=None,
            anomalyCount=None,
            slaBreachCount=None,
            executionStatus=None,
            ruleExecutionType=None,
            *args,
            **kwargs
    ):
        # Initialize base class
        super().__init__(status=status, ruleExecutionType=ruleExecutionType)

        # Freshness-specific fields
        self.description = description
        self.successCount = successCount
        self.failureCount = failureCount
        self.qualityScore = qualityScore
        self.anomalyCount = anomalyCount
        self.slaBreachCount = slaBreachCount
        self.executionStatus = executionStatus

    def __repr__(self):
        return f"FreshnessRuleExecutionResult({self.__dict__})"

# --------------------------------------------------------------------
# FreshnessRule (aligned to root['rule'])
# --------------------------------------------------------------------
class FreshnessRule:
    def __init__(
            self,
            id=None,
            name=None,
            description=None,
            notificationChannels=None,
            enabled=None,
            backingAsset=None,
            createdAt=None,
            updatedAt=None,
            thresholdLevel=None,
            archived=None,
            archivalReason=None,
            tenantId=None,
            createdBy=None,
            lastUpdatedBy=None,
            tags=None,
            sparkResourceConfig=None,
            policyGroups=None,
            version=None,
            type=None,
            autoCreated=False,
            anomalyStrengthThreshold=None,
            resourceStrategyType=None,
            autoRetryEnabled=False,
            selectedResourceInventory=None,
            includeInQualityScore=None,
            schedule=None,
            scheduled=None,
            timeZone="GMT",
            timeSecondsOffset=None,
            jobSchedule=None,
            *args,
            **kwargs,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.notificationChannels = notificationChannels
        self.enabled = enabled

        self.backingAsset = (
            BackingAsset(**backingAsset)
            if isinstance(backingAsset, dict)
            else backingAsset
        )

        self.createdAt = createdAt
        self.updatedAt = updatedAt

        self.thresholdLevel = (
            parse_rule_threshold_configuration(thresholdLevel)
            if isinstance(thresholdLevel, dict)
            else thresholdLevel
        )

        self.archived = archived
        self.archivalReason = archivalReason
        self.tenantId = tenantId
        self.createdBy = createdBy
        self.lastUpdatedBy = lastUpdatedBy

        self.tags = [
            RuleTag(**obj) if isinstance(obj, dict) else obj
            for obj in (tags or [])
        ]

        self.sparkResourceConfig = (
            SparkResourceConfig(**sparkResourceConfig)
            if isinstance(sparkResourceConfig, dict)
            else sparkResourceConfig
        )

        self.policyGroups = [
            PolicyGroup(**obj) if isinstance(obj, dict) else obj
            for obj in (policyGroups or [])
        ]

        self.version = version
        self.type = type
        self.autoCreated = autoCreated
        self.anomalyStrengthThreshold = anomalyStrengthThreshold
        self.resourceStrategyType = resourceStrategyType
        self.autoRetryEnabled = autoRetryEnabled
        self.selectedResourceInventory = selectedResourceInventory
        self.includeInQualityScore = includeInQualityScore
        self.schedule = schedule
        self.scheduled = scheduled
        self.timeZone = timeZone
        self.timeSecondsOffset = timeSecondsOffset
        self.jobSchedule = jobSchedule

    def __repr__(self):
        return f"FreshnessRule({self.__dict__})"


# --------------------------------------------------------------------
# FreshnessRuleDetails (aligned to root['details'])
# --------------------------------------------------------------------
class FreshnessRuleDetails:
    def __init__(
            self,
            id=None,
            ruleId=None,
            backingAssetId=None,
            items=None,
            isSegmented=False,
            segments=None,
            isCompositeRule=False,
            continueExecutionOnFailure=False,
            version=None,
            *args,
            **kwargs,
    ):
        self.id = id
        self.ruleId = ruleId
        self.backingAssetId = backingAssetId
        self.isSegmented = isSegmented
        self.isCompositeRule = isCompositeRule
        self.continueExecutionOnFailure = continueExecutionOnFailure
        self.version = version

        # Items list — dict → Item conversion
        self.items = [Item(**obj) if isinstance(obj, dict) else obj for obj in (items or [])]

        # Segments list
        self.segments = [
            ShortSegment(**obj) if isinstance(obj, dict) else obj
            for obj in (segments or [])
        ]

    def __repr__(self):
        return f"FreshnessRuleDetails({self.__dict__})"


# --------------------------------------------------------------------
# FreshnessRuleResource (wrapper + executor)
# --------------------------------------------------------------------
class FreshnessRuleResource(RuleResource):
    def __init__(self, rule, details, client=None, *args, **kwargs):
        self.rule = FreshnessRule(**rule) if isinstance(rule, dict) else rule
        self.details = FreshnessRuleDetails(**details) if isinstance(details, dict) else details
        self.client = client

    def get_executions(self, page=0, size=25, sortBy='finishedAt:DESC'):
        return self.client.policy_executions(self.rule.id, PolicyType.DATA_CADENCE, page, size, sortBy)

    def execute(
            self,
            sync=True,
            incremental=False,
            failure_strategy: FailureStrategy = FailureStrategy.DoNotFail,
            policy_execution_request: Optional[PolicyExecutionRequest] = None,
    ) -> RuleExecution:
        logger.info(
            f"Policy Execution Request: "
            f"{policy_execution_request.to_dict() if policy_execution_request else 'None'}"
        )
        return self.client.execute_rule(
            PolicyType.DATA_CADENCE,
            self.rule.id,
            sync,
            incremental,
            failure_strategy,
            policy_execution_request,
        )

    def __repr__(self):
        return f"FreshnessRuleResource({self.__dict__})"



from dataclasses import dataclass
from enum import Enum
from datetime import datetime


@dataclass
class SpanMetadata:
    owner: str = None


class SpanStatus(Enum):
    INITIALIZED = 1
    STARTED = 2
    FINISHED = 3
    ABORTED = 4
    FAILED = 5


class Span:

    def __init__(self, uid, pipelineRunId, status=SpanStatus.INITIALIZED, parentSpanId=None, id=None, startedAt=None,
                 finishedAt=None, *args, **kwargs):
        self.uid = uid
        self.pipelineRunId = pipelineRunId
        self.status = status
        if parentSpanId is not None:
            self.parentSpanId = parentSpanId
        if id is not None:
            self.id = id
            self.startedAt = startedAt
            self.finishedAt = finishedAt

    def __eq__(self, other):
        return self.uid == other.uid

    def __repr__(self):
        return f"Span({self.__dict__!r})"


class SpanEventType(Enum):
    SPAN_START = 'SPAN.START'
    SPAN_END = 'SPAN.END'
    SPAN_ABORTED = 'SPAN.ABORTED'
    SPAN_FAILED = 'SPAN.FAILED'


class SpanContextEvent:

    def __init__(self, eventUid: SpanEventType, spanId: int, contextData=None, id=None, *args, **kwargs):
        """
            Description:
                span context event
        :param eventUid: event type
        :param spanId: span id
        :param contextData: context data of the event
        :param id: id of the event
        """
        self.eventUid = eventUid
        self.spanId = spanId
        self.contextData = contextData
        if id is not None:
            self.id = id

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"SpanContextEvent({self.__dict__!r})"


class CreateSpanEvent:

    def __init__(self, event_uid: str, span_id: int, context_data: dict = None, log_data: str = None,
                 created_at: datetime = None, *args, **kwargs):
        """
            Description:
                Class used to create event for the span
            :param event_uid:
            :param span_id:
            :param context_data:
            :param log_data
            :param created_at
        """
        self.eventUid = event_uid
        self.contextData = context_data
        self.logData = log_data
        self.spanId = span_id
        self.createdAt = created_at

    def __repr__(self):
        return f"CreateSpanEvent({self.__dict__!r})"



from acceldata_sdk.events.span_event import SpanEvent
from datetime import datetime

class GenericEvent(SpanEvent):
    """
        Description:
            Class to send any custom event to ADOC. `event_uid` should be set
            here and the context data.
    """

    def __init__(self, event_uid, context_data=None, created_at: datetime = None):
        """
        :param event_uid: The unique identifier for the event.
        :param context_data: Optional context data for the event
        :param created_at: An optional datetime representing the explicit time for sending the generic event.
         If not set, the current time will be used to send the generic event
        """
        self.context_data = context_data
        self.event_uid = event_uid
        self.created_at = created_at
        self.created_at = created_at



import logging
import pytest
from datetime import datetime
import random

from acceldata_sdk.models.pipeline import (
    CreatePipeline,
    PipelineMetadata,
    PipelineRunResult,
    PipelineRunStatus,
)
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.events.generic_event import GenericEvent

from . import test_constants as test_const

logger = logging.getLogger(__name__)


@pytest.mark.integration
def test_list_all_pipelines(adoc_client):
    logger.info("Listing all pipelines")
    pipelines = adoc_client.get_pipelines()
    assert pipelines
    logger.info("Fetched %d pipelines", len(pipelines))


@pytest.mark.integration
def test_create_pipeline(adoc_client):
    logger.info("Creating pipeline")

    meta = PipelineMetadata(
        owner="sdk/pipeline-user",
        team="ADOC",
        codeLocation="...",
    )

    pipeline = CreatePipeline(
        uid=test_const.PIPELINE_UID,
        name=test_const.PIPELINE_NAME,
        description="Pipeline created using Acceldata SDK",
        meta=meta,
        context={
            "pipeline_uid": test_const.PIPELINE_UID,
            "pipeline_name": test_const.PIPELINE_NAME,
        },
    )

    pipeline_res = adoc_client.create_pipeline(pipeline=pipeline)
    assert pipeline_res is not None

    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    assert pipeline is not None

    logger.info("Pipeline created successfully")


@pytest.mark.integration
def test_create_pipeline_run(adoc_client):
    logger.info("Creating pipeline run")

    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    pipeline_run = pipeline.create_pipeline_run()

    fetched_run = pipeline.get_run(pipeline_run.id)
    assert fetched_run is not None
    assert pipeline_run.id == fetched_run.id

    logger.info("Pipeline run created with id=%s", pipeline_run.id)


@pytest.mark.integration
def test_latest_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)

    latest_run = pipeline.get_latest_pipeline_run()
    runs = pipeline.get_runs()

    assert latest_run.id == runs[0].id
    logger.info("Latest pipeline run verified")


@pytest.mark.integration
def test_create_root_span(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    span = run.create_span(uid=f"{test_const.PIPELINE_UID}.root.span")
    assert span is not None

    logger.info("Root span created")


@pytest.mark.integration
def test_update_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    result = run.update_pipeline_run(
        context_data={"name": "backend"},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED,
    )

    assert result is not None
    logger.info("Pipeline run updated")


@pytest.mark.integration
def test_pipeline_run_details(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    details = run.get_details()
    assert details is not None

    logger.info("Fetched pipeline run details")


@pytest.mark.integration
def test_create_job(adoc_client):
    logger.info("Creating job")

    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    job = CreateJob(
        uid=test_const.PIPELINE_JOB_UID,
        name=f"{test_const.PIPELINE_JOB_UID} Job",
        pipeline_run_id=run.id,
        description="Job created using torch SDK",
        inputs=[Node(job_uid="customers.data-generation")],
        outputs=[Node(asset_uid="S3-DS.s3_customers")],
        meta=JobMetadata(
            "ADOC Pipeline",
            "DR Team",
            "https://github.com/coke/reports/customers.kt",
        ),
        context={
            "job": "data_gene",
            "operator": "write_file_func",
            "time": str(datetime.now()),
        },
    )

    job_res = pipeline.create_job(job)
    assert job_res is not None

    logger.info("Job created successfully")


@pytest.mark.integration
def test_create_child_span(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    root_span = run.get_root_span()
    child_span = root_span.create_child_span(
        uid=f"{test_const.PIPELINE_JOB_UID}.span",
        context_data={"time": str(datetime.now())},
        associatedJobUids=[test_const.PIPELINE_JOB_UID],
    )

    child_span.start()
    child_span.send_event(
        GenericEvent(
            context_data={
                "total_file": 1,
                "schema": "name,address,dept_id",
            },
            event_uid="customers.data.generation.metadata",
        )
    )
    child_span.end()
    root_span.end()

    spans = root_span.get_child_spans()
    assert spans
    logger.info("Child span created and validated")


@pytest.mark.integration
def test_spans_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    spans = run.get_spans()
    assert spans is not None

    logger.info("Fetched pipeline spans")

@pytest.mark.integration
def test_create_pipeline_run_with_continuation_id(adoc_client):
    create_pipeline_payload = CreatePipeline(
        uid=test_const.PIPELINE_UID_WITH_CONTINUATION_ID,
        name=test_const.PIPELINE_NAME_WITH_CONTINUATION_ID,
        description=f'The has been created from sdk',
    )
    pipeline = adoc_client.create_pipeline(pipeline=create_pipeline_payload)
    logger.info("Validating test_create_pipeline_run_with_continuation_id")
    continuation_id=f'run.{random.random()}'
    pipeline_run = pipeline.create_pipeline_run(continuation_id=continuation_id)
    logger.info(pipeline_run)
    run_with_id = pipeline.get_run(continuation_id=continuation_id)
    logger.info(run_with_id)
    assert pipeline_run == run_with_id

@pytest.mark.integration
def test_delete_pipeline(adoc_client):

    logger.info("Creating pipeline")

    meta = PipelineMetadata(
        owner="sdk/pipeline-user",
        team="ADOC",
        codeLocation="...",
    )

    pipeline = CreatePipeline(
        uid=test_const.PIPELINE_UID_DELETION,
        name=test_const.PIPELINE_NAME_DELETION,
        description="Pipeline created using Acceldata SDK",
        meta=meta,
        context={
            "pipeline_uid": test_const.PIPELINE_UID_DELETION,
            "pipeline_name": test_const.PIPELINE_NAME_DELETION,
        },
    )

    pipeline_res = adoc_client.create_pipeline(pipeline=pipeline)
    res = pipeline_res.delete()

    assert res is not None
    logger.info("Pipeline deleted successfully")



import logging

import pytest
from acceldata_sdk.errors import TorchSdkException, APIError
from acceldata_sdk.models.common_types import (
    ExecutionType,
    BoundsIdMarkerConfig,
    BoundsDateTimeMarkerConfig,
    BoundsFileEventMarkerConfig,
    TimestampBasedMarkerConfig,
)
from acceldata_sdk.models.profile import StartProfilingRequest

from . import test_constants as test_const
from ..commons.retry import retry_operation

logger = logging.getLogger(__name__)

def execute_profiling_or_accept_running(operation):
    try:
        result = retry_operation(
            operation,
            test_const.MAX_RETRIES,
            test_const.RETRY_INTERVAL,
        )
        return result

    except APIError as e:
        msg = str(e)
        logger.info(msg)

        # ACCEPTABLE: profiling already running
        if (
                "profile is already running" in msg
                or "Cannot start FULL profiling" in msg
                or "409" in msg
        ):
            logger.info("Profiling already running – accepting as PASS")
            return "Profiling already running"

        # ACCEPTABLE: previous execution not completed
        if (
                "Previous execution" in msg
                and "has not completed" in msg
        ):
            logger.info("Previous profiling execution still running – accepting as PASS")
            return "Previous profiling execution still running"

        # Anything else is a real failure
        raise


@pytest.fixture(scope="module")
def table_asset(adoc_client):
    return adoc_client.get_asset(test_const.TABLE_ASSET_UID)


@pytest.mark.integration
def test_execute_full_profiling_backward_compatible(table_asset):
    logger.info("Executing FULL profiling (backward compatible)")

    def operation():
        return table_asset.start_profile(
            profiling_type=ExecutionType.FULL
        )

    result = execute_profiling_or_accept_running(operation)
    logger.info("Profiling result: %s", result)

    # PASS if started or already running
    assert result is None or result is not None


@pytest.mark.integration
def test_execute_incremental_profiling_backward_compatible(table_asset):
    logger.info("Executing INCREMENTAL profiling (backward compatible)")

    def operation():
        return table_asset.start_profile(
            profiling_type=ExecutionType.INCREMENTAL
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_full_profiling_using_request(table_asset):
    request = StartProfilingRequest(
        profilingType=ExecutionType.FULL
    )

    logger.info("Executing FULL profiling using StartProfilingRequest")

    def operation():
        return table_asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_incremental_profiling_using_request(table_asset):
    request = StartProfilingRequest(
        profilingType=ExecutionType.INCREMENTAL
    )

    logger.info("Executing INCREMENTAL profiling using StartProfilingRequest")

    def operation():
        return table_asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_selective_profiling_without_marker_config(table_asset):
    request = StartProfilingRequest(
        profilingType=ExecutionType.SELECTIVE
    )

    logger.info("Executing SELECTIVE profiling without marker config")

    with pytest.raises(TorchSdkException):
        table_asset.start_profile(
            start_profiling_request=request
        )


@pytest.mark.integration
def test_execute_selective_profiling_id_based(table_asset):
    marker = BoundsIdMarkerConfig(
        idColumnName="ID",
        fromId=0,
        toId=1000,
    )

    request = StartProfilingRequest(
        profilingType=ExecutionType.SELECTIVE,
        markerConfig=marker,
    )

    logger.info("Executing SELECTIVE profiling (ID based)")

    def operation():
        return table_asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_selective_profiling_datetime_based(table_asset):
    marker = BoundsDateTimeMarkerConfig(
        dateColumnName="TO_DATE",
        format="yyyy-MM-dd",
        fromDate="2023-07-01 00:00:00.000",
        toDate="2024-07-14 23:59:59.999",
        timeZoneId="Asia/Calcutta",
    )

    request = StartProfilingRequest(
        profilingType=ExecutionType.SELECTIVE,
        markerConfig=marker,
    )

    logger.info("Executing SELECTIVE profiling (datetime based)")

    def operation():
        return table_asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_selective_profiling_file_event_based(adoc_client):
    asset = adoc_client.get_asset(
        test_const.FILE_BASED_ASSET_UID
    )

    marker = BoundsFileEventMarkerConfig(
        fromDate="2019-04-01 00:00:00.000",
        toDate="2024-07-16 23:59:59.999",
        timeZoneId="Asia/Calcutta",
    )

    request = StartProfilingRequest(
        profilingType=ExecutionType.SELECTIVE,
        markerConfig=marker,
    )

    logger.info("Executing SELECTIVE profiling (file event based)")

    def operation():
        return asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None

@pytest.mark.integration
def test_execute_selective_profiling_kafka_timestamp_based(adoc_client):
    asset = adoc_client.get_asset(
        test_const.KAFKA_ASSET_UID
    )

    marker = TimestampBasedMarkerConfig(
        format="yyyy-mm-dd",
        initialOffset="2023-06-01",
        timeZoneId="Asia/Calcutta",
    )

    request = StartProfilingRequest(
        profilingType=ExecutionType.SELECTIVE,
        markerConfig=marker,
    )

    logger.info("Executing SELECTIVE profiling (kafka timestamp based)")

    def operation():
        return asset.start_profile(
            start_profiling_request=request
        )

    result = execute_profiling_or_accept_running(operation)

    logger.info("Profiling result: %s", result)
    assert result is not None



import time
import pytest
import logging
from acceldata_sdk.errors import APIError

LOGGER = logging.getLogger(__name__)


def retry_operation(operation, max_retries, retry_interval, *args, **kwargs):
    retry_count = 0
    LOGGER.info("Retry count: %s", retry_count)
    while retry_count < max_retries+1:
        try:
            LOGGER.info("Executing operation...")
            result = operation(*args, **kwargs)
            LOGGER.info(f"Operation result: {result}")
            return result
        except Exception as e:
            LOGGER.error(str(e))

            # ✅ Non-retriable / acceptable API states → rethrow
            if isinstance(e, APIError):
                raise

            retry_count += 1
            if retry_count < max_retries:
                time.sleep(retry_interval)
            else:
                pytest.fail("Operation failed after multiple retries.")
    return None



import logging
import pprint

import pytest
from acceldata_sdk.errors import APIError
from acceldata_sdk.models.profile import ProfilingType
from acceldata_sdk.models.tags import AssetLabel, CustomAssetMetadata

from . import test_constants as test_const
from ..commons.retry import retry_operation

pp = pprint.PrettyPrinter(indent=4)

logger = logging.getLogger()


@pytest.mark.integration
def test_get_datasource(adoc_client):
    datasource_response = adoc_client.get_datasource(test_const.DS_NAME, True)
    logger.info("Datasource response: %s", datasource_response)
    assert datasource_response is not None


@pytest.mark.integration
def test_get_datasource_id(adoc_client):
    datasource_response = adoc_client.get_datasource(test_const.DS_NAME, True)
    datasource_by_id_response = adoc_client.get_datasource(datasource_response.id, False)
    logger.info("Datasource by ID response: %s", datasource_by_id_response)
    assert datasource_by_id_response is not None


@pytest.mark.integration
def test_get_all_data_sources(adoc_client):
    all_data_sources = adoc_client.get_datasources()
    logger.info("All datasources response:")
    logger.info(pp.pformat(all_data_sources))
    assert all_data_sources is not None


@pytest.mark.integration
def test_get_ds_crawler(adoc_client):
    datasource = adoc_client.get_datasource(test_const.DS_NAME, False)
    ds_id = adoc_client.get_datasource(datasource.id, True)
    try:
        status = ds_id.get_crawler_status()
        logger.info("Crawler status response:")
        logger.info(pp.pformat(status))
    except APIError as e:
        msg = str(e)
        logger.info(msg)

        # ACCEPTABLE outcome
        if "Unable to start the Crawler" in msg or "422" in msg:
            assert True
            return

        # Anything else is a real failure
        raise
    assert status is not None


@pytest.mark.integration
def test_start_crawler(adoc_client):
    datasource = adoc_client.get_datasource(test_const.DS_NAME, False)

    try:
        datasource.start_crawler()
        # If no exception → crawler started → PASS
        assert True

    except APIError as e:
        msg = str(e)
        logger.info(msg)

        # ACCEPTABLE outcome
        if "Unable to start the Crawler" in msg or "422" in msg:
            assert True
            return

        # Anything else is a real failure
        raise


@pytest.mark.integration
def test_get_asset(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    logger.info("Asset by UID response: %s", asset)
    assert asset is not None


@pytest.mark.integration
def test_get_asset_id(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    asset = adoc_client.get_asset(asset.id)
    logger.info("Asset by ID response: %s", asset)
    assert asset is not None


@pytest.mark.integration
def test_get_asset_metadata(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    metadata = asset.get_metadata()
    logger.info("Asset metadata response:")
    logger.info(pp.pformat(metadata))
    assert metadata is not None


@pytest.mark.integration
def test_get_asset_sample_data(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    sample_data = asset.sample_data()
    logger.info("Sample data response:")
    logger.info(pp.pformat(sample_data))
    assert sample_data is not None


@pytest.mark.integration
def test_get_asset_labels(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    labels = asset.get_labels()
    logger.info("Asset labels response:")
    logger.info(pp.pformat(labels))
    assert labels is not None


@pytest.mark.integration
def test_add_asset_labels(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)

    response = asset.add_labels(
        labels=[
            AssetLabel("test12", "shubh12"),
            AssetLabel("test22", "shubh32"),
        ]
    )

    logger.info("Add labels response: %s", response)

    labels = asset.get_labels()
    logger.info("Labels after add:")
    logger.info(pp.pformat(labels))

    assert labels is not None


@pytest.mark.integration
def test_add_asset_custom_metadata(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)

    response = asset.add_custom_metadata(
        custom_metadata=[
            CustomAssetMetadata("testcm1", "shubhcm1"),
            CustomAssetMetadata("testcm2", "shubhcm2"),
        ]
    )

    logger.info("Add custom metadata response: %s", response)

    metadata = asset.get_metadata()
    logger.info("Metadata after add:")
    logger.info(pp.pformat(metadata))

    assert metadata is not None


@pytest.mark.integration
def test_profile_status(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    status = asset.get_latest_profile_status()
    logger.info("Latest profile status:")
    logger.info(pp.pformat(status))
    assert status is not None


@pytest.mark.integration
def test_add_tag(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)

    try:
        asset.add_tag(test_const.ASSET_TAG)
    except APIError as e:
        assert "already exists" in str(e)


@pytest.mark.integration
def test_cancel_profile(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)
    def operation():
        return asset.start_profile(ProfilingType.FULL)

    try:
        profile = retry_operation(
            operation,
            test_const.MAX_RETRIES,
            test_const.RETRY_INTERVAL
        )
    except APIError as e:
        msg = str(e)

        # ✅ ACCEPTABLE: already running
        if "A profile is already running" in msg or "409" in msg:
            logger.info("Profile already running — accepting as PASS")
            return

        raise

    logger.info(f"Profile response: {profile}")
    status = profile.get_status()
    logger.info("Profile status before cancel:")
    logger.info(pp.pformat(status))

    if status["profileRequest"]["status"] == "IN PROGRESS":
        cancel_res = profile.cancel()
        logger.info("Cancel profile response: %s", cancel_res)
        assert cancel_res is not None


@pytest.mark.integration
def test_execute_profile(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID)

    def operation():
        return asset.start_profile(ProfilingType.FULL)

    try:
        profile = retry_operation(
            operation,
            test_const.MAX_RETRIES,
            test_const.RETRY_INTERVAL
        )
    except APIError as e:
        msg = str(e)

        # ✅ ACCEPTABLE: already running
        if "A profile is already running" in msg or "409" in msg:
            logger.info("Profile already running — accepting as PASS")
            return

        raise
    status = profile.get_status()
    logger.info("Profile execution status:")
    logger.info(pp.pformat(status))
    status = profile.get_status()
    logger.info("Profile execution status:")
    logger.info(pp.pformat(status))

    assert status is not None



from enum import Enum
from typing import List, Optional, Dict, Any
from acceldata_sdk.models.common_types import ExecutionType, MarkerConfig

class Profile:

    def __init__(self, assetId, createdAt, id, jobId=None, profilingType=None, sparkClusterContext=None, status=None,
                 updatedAt=None, rows=None,
                 totalRows=None, jobType=None, autoProfileId=None, profileDataAvailable=None, marker=None,
                 profiledData=None, error=None, client=None, *args, **kwargs):
        self.client = client
        self.autoProfileId = autoProfileId
        self.profileDataAvailable = profileDataAvailable
        self.jobType = jobType
        self.updatedAt = updatedAt
        self.status = status
        self.sparkClusterContext = sparkClusterContext
        self.rows = rows
        self.profilingType = profilingType
        self.profiledData = profiledData
        self.marker = marker
        self.jobId = jobId
        self.id = id
        self.error = error
        self.createdAt = createdAt
        self.assetId = assetId
        self.totalRows = totalRows

    def __repr__(self):
        return f"Profile({self.__dict__})"

    def cancel(self):
        return self.client.cancel_profile(self.id)

    def get_status(self):
        return self.client.get_profile_request_details(asset_id=self.assetId, req_id=self.id)

class ProfilingType(Enum):
    SAMPLE = 'SELECTIVE'
    FULL = 'FULL'
    INCREMENTAL = 'INCREMENTAL'


class JobType(Enum):
    PROFILE = 'profile'
    MINI_PROFILE = 'mini-profile'
    # FULL = 'autotag'


class ProfileRequest:

    def __init__(self, id, status, createdAt=None, updatedAt=None, client=None, totalRows=None, isProfileAnomalous=None,
                 *args, **kwargs):
        self.client = client
        self.updatedAt = updatedAt
        self.createdAt = createdAt
        self.status = status
        self.id = id
        self.totalRows = totalRows
        self.isProfileAnomalous = isProfileAnomalous

    def __repr__(self):
        return f"ProfileRequest({self.__dict__})"

    def cancel(self):
        return self.client.cancel_profile(self.id)


class StartProfilingRequest:
    def __init__(self, profilingType: 'ExecutionType', markerConfig: 'MarkerConfig' = None, *args, **kwargs):
        """
        Initializes a StartProfilingRequest object.

        Args:
            profilingType (ExecutionType): The type of profiling to be executed.
            markerConfig (MarkerConfig, optional): The configuration for the markers. Defaults to None.
        """
        self.profilingType = profilingType
        self.markerConfig = markerConfig

    def to_dict(self) -> Dict[str, Any]:
        return {
            "data": {
                "profilingType": self.profilingType.to_dict(),
                "markerConfig": self.markerConfig.to_dict() if self.markerConfig else None
            }
        }



from enum import Enum

class ThresholdType(Enum):
    ANOMALY = "ANOMALY"
    RELATIVE = "RELATIVE"
    TIME_BASED = "TIME_BASED"
    ABSOLUTE = "ABSOLUTE"

    def display_name(self):
        return self.name


class RuleThresholdConfiguration:
    @property
    def type(self) -> ThresholdType:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__})"


class AbsoluteThresholdConfig(RuleThresholdConfiguration):
    def __init__(self, config):
        self.config = config

    @property
    def type(self):
        return ThresholdType.ABSOLUTE


class AnomalyThresholdConfig(RuleThresholdConfiguration):
    def __init__(self, config):
        self.config = config

    @property
    def type(self):
        return ThresholdType.ANOMALY


class RelativeThresholdConfig(RuleThresholdConfiguration):
    def __init__(self, config, operational_config):
        self.config = config
        self.operational_config = operational_config

    @property
    def type(self):
        return ThresholdType.RELATIVE


class TimeBasedThresholdConfig(RuleThresholdConfiguration):
    def __init__(self, config):
        self.config = config

    @property
    def type(self):
        return ThresholdType.TIME_BASED


RuleLevelThresholdConfiguration = RuleThresholdConfiguration

def parse_rule_threshold_configuration(data: dict) -> RuleThresholdConfiguration:
    if not data:
        return None

    threshold_type = data.get("type")

    if threshold_type == "ABSOLUTE":
        return AbsoluteThresholdConfig(config=data.get("config"))

    if threshold_type == "ANOMALY":
        return AnomalyThresholdConfig(config=data.get("config"))

    if threshold_type == "RELATIVE":
        return RelativeThresholdConfig(
            config=data.get("config"),
            operational_config=data.get("operationalConfig"),
        )

    if threshold_type == "TIME_BASED":
        return TimeBasedThresholdConfig(config=data.get("config"))

    raise ValueError(f"Unsupported ThresholdType: {threshold_type}")




# ============================================================
# DATASOURCE CONSTANTS
# ============================================================

DS_NAME = "sdk_aws_ds"
DS_ID = 84902

# ============================================================
# ASSET CONSTANTS
# ============================================================

ASSET_UID = "sdk_snowflake_ds.SDK_DB.PUBLIC.CLIENTLIST_SOURCEAPP"

ASSET_TAG = "sdk_asset"

ASSET_UID_TABLE = "sdk_snowflake_ds.SDK_DB.POSITIONS"

# ============================================================
# PIPELINE CONSTANTS
# ============================================================

PIPELINE_UID = "torch.external.integration.demo-default"
PIPELINE_NAME = "SDK External Integration Pipeline - Default"

PIPELINE_UID_WITH_CONTINUATION_ID = "adoc.test.pipeline.sanity.continuation"
PIPELINE_NAME_WITH_CONTINUATION_ID = "ADOC Test Pipeline Sanity ContinuationId"

PIPELINE_UID_DELETION = "adoc_pipeline_to_delete"
PIPELINE_NAME_DELETION = "ETL Pipeline To Delete"

PIPELINE_UID_CONTINUATION = "adoc.test.pipeline.sanity.continuation"

PIPELINE_JOB_UID = "read_data_from_s3"

JOB_UID_READ = "customers.read"
JOB_UID_GENERATE = "customers.generate_sales"
JOB_UID_SALES = "aggregated_sales"

# ============================================================
# EXPLICIT TIME PIPELINE CONSTANTS
# ============================================================

EXPLICIT_PIPELINE_UID = "torch.external.integration.time-travel"
EXPLICIT_PIPELINE_NAME = "External Integration – Time Travel (Explicit Time)"

# ============================================================
# EXTERNAL INTEGRATION CONSTANTS
# ============================================================

S3_DS = "sdk_aws_ds"
SDK_SNOWFLAKE_DATA_SOURCE = "sdk_snowflake_ds"

S3_CUSTOMER = "customers_raw"
S3_SNOWFLAKE_CUSTOMERS = "SDK_DB.PUBLIC.TELCO_CUSTOMER_NETWORK_SCORE"
SNOWFLAKE_SERVICES = "SDK_DB.PUBLIC.TMOBILE_SERVICESBYCUSTOMER_V2"

# ============================================================
# POLICY CONSTANTS
# ============================================================

# -------------------------
# Freshness
# -------------------------

FRESHNESS_POLICY_NAME = (
    "SALES_PIPELINE_CLEAN-fresh-and-vol-policy-77428679"
)

# -------------------------
# Data Quality
# -------------------------

DQ_POLICY_NAME = "sdk_sko_full_dq_policy"

DQ_POLICY_BACKWARD_COMPATIBLE_NAME = "sdk_full_telecom_policy"
INCREMENTAL_DQ_POLICY_NAME = "sdk_erp_transactions_incremental"

INCREMENTAL_DQ_ID = "sdk_semistructured_dq_incremental_id"
INCREMENTAL_DQ_DATE_TIME="sdk_sales_pipeline_dq_date_incremental"
INCREMENTAL_DQ_DATE_TIME_ASSET_ID=77428679

FILE_EVENT_BASED_DQ_POLICY = "sdk_customers_json_file_event_based_dq"
FILE_EVENT_BASED_DQ_BACKING_ASSET_ID=77429408

KAFKA_DQ_POLICY_NAME = "sdk_kafka_timestamp_dq_policy"
KAFKA_DQ_POLICY_BACKING_ASSET_ID=77429436

# -------------------------
# Reconciliation
# -------------------------

RECON_POLICY_NAME = "sdk_sales_recon_policy"

SELECTIVE_RECON_POLICY_BACKWARD_COMPATIBLE_NAME = RECON_POLICY_NAME
INCREMENTAL_RECON_POLICY_NAME = "sdk_sales_incremental_date_recon_policy"
INCREMENTAL_DQ_RECON_TIME_ASSET_ID=77428678

FILE_EVENT_BASED_RECON_POLICY = "file_event_based_recon_policy"

KAFKA_RECON_POLICY_NAME = "sdk_kafka_timestamp_recon_policy"

FILE_BASED_ASSET_UID = "sdk_aws_ds.customers_raw"
KAFKA_ASSET_UID = "sdk_kafka.sdk_cust_sink_kafka"
TABLE_ASSET_UID = "sdk_snowflake_ds.SDK_DB.PUBLIC.CLINET_DEMOGRAPHIC"

# ============================================================
# CSV / DATA GENERATION
# ============================================================

CSV_SIZE = 30

# ============================================================
# RETRY SETTINGS
# ============================================================

RETRY_INTERVAL = 60
MAX_RETRIES = 0



# acceldata_sdk/utils/parsing.py

def str_to_bool(value: str) -> bool:
    """
    Convert a string representation of truth to boolean.

    Accepted true values:
        '1', 'true', 'yes', 'y', 'on'

    Accepted false values:
        '0', 'false', 'no', 'n', 'off'

    Raises:
        ValueError if value is not a valid boolean string.
    """
    val = value.strip().lower()

    if val in {"1", "true", "yes", "y", "on"}:
        return True
    if val in {"0", "false", "no", "n", "off"}:
        return False

    raise ValueError(f"Invalid boolean value: {value}")



from acceldata_sdk.models.span import CreateSpanEvent
from acceldata_sdk.utils.datetime_utils import DateTimeFormatter
from datetime import datetime


class SpanEvent:
    """
        Description:
            Base class for all span event supported in torch
    """

    def event_uid_validation(self, event_uid: str):
        uid = event_uid.replace(" ", "")
        return uid.replace("_", ".")

    def convert(self, event_uid: str, span_id: int, context_data=None, log_data=None, client=None,
                created_at: datetime = None):
        """
        :param event_uid: event uid. similar to event type
        :param log_data: log data for the event
        :param context_data: optional context data map
        :param span_id: span id
        :param client: TorchClient class instance
        :param created_at: An optional datetime representing the explicit time for sending the span event.
         If not set, the current time will be used to send the span event
        :return: SpanContextEvent class instance of the executed event
        """
        event_uid = self.event_uid_validation(event_uid)
        created_at = DateTimeFormatter.format_datetime(created_at)
        span_event = CreateSpanEvent(
            event_uid=event_uid,
            span_id=span_id,
            context_data=context_data,
            log_data=log_data,
            created_at=created_at
        )
        event_response = client.create_span_event(span_event)
        return event_response



import logging
import pprint
import pytest
import acceldata_sdk.constants as const

from acceldata_sdk.models.ruleExecutionResult import (
    RuleType,
    PolicyFilter,
)

from . import test_constants as test_const
from ..commons.retry import retry_operation

pp = pprint.PrettyPrinter(indent=4)
logger = logging.getLogger()


# ============================================================
# DQ POLICY TESTS
# ============================================================

@pytest.mark.integration
def test_get_dq_policy(adoc_client):
    policy = adoc_client.get_policy(
        const.PolicyType.DATA_QUALITY,
        test_const.DQ_POLICY_NAME,
    )

    logger.info("DQ policy response:")
    logger.info(pp.pformat(policy))

    assert policy is not None


@pytest.mark.integration
def test_get_dq_policy_without_type(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.DATA_QUALITY,
        test_const.DQ_POLICY_NAME,
    )

    policy_by_id = adoc_client.get_policy(identifier=policy_response.rule.id)

    logger.info("DQ policy without type:")
    logger.info(pp.pformat(policy_by_id))

    assert policy_by_id is not None


@pytest.mark.integration
def test_list_all_dq_policies(adoc_client):
    filter_ = PolicyFilter(
        policyType=RuleType.DATA_QUALITY,
        enable=True,
    )

    policies = adoc_client.list_all_policies(filter=filter_)

    logger.info("DQ policies list:")
    logger.info(pp.pformat(policies))

    assert policies is not None


@pytest.mark.integration
def test_execute_dq_policy_async(adoc_client):
    policy = adoc_client.get_policy(
        type=const.PolicyType.DATA_QUALITY,
        identifier=test_const.DQ_POLICY_NAME,
    )

    def operation():
        return policy.execute(sync=False)

    executor = retry_operation(
        operation,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL,
    )

    if executor.errorMessage:
        pytest.skip(executor.errorMessage)

    status = executor.get_status()
    logger.info("DQ execution status:")
    logger.info(pp.pformat(status))

    result = executor.get_result()
    logger.info("DQ execution result:")
    logger.info(pp.pformat(result))

    assert result is not None


@pytest.mark.integration
def test_cancel_dq_policy(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.DATA_QUALITY,
        test_const.DQ_POLICY_NAME,
    )

    executor = policy_response.execute(sync=False)
    logger.info("executor:")
    logger.info(pp.pformat(executor))
    if executor.errorMessage:
        pytest.skip("Execution did not start")

    executor.cancel()
    rule_result = executor.get_result(executor.id)
    logger.info("DQ cancel response:")
    logger.info(pp.pformat(rule_result))

    assert rule_result is not None


@pytest.mark.integration
def test_get_dq_policy_executions(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.DATA_QUALITY,
        test_const.DQ_POLICY_NAME,
    )

    executions = adoc_client.policy_executions(
        policy_response.rule.id,
        RuleType.DATA_QUALITY,
    )

    logger.info("DQ executions:")
    logger.info(pp.pformat(executions))

    assert executions is not None


# ============================================================
# RECON POLICY TESTS
# ============================================================

@pytest.mark.integration
def test_get_recon_policy(adoc_client):
    policy = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )

    logger.info("Recon policy response:")
    logger.info(pp.pformat(policy))

    assert policy is not None


@pytest.mark.integration
def test_get_recon_policy_without_type(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )

    policy_by_id = adoc_client.get_policy(identifier=policy_response.rule.id)

    logger.info("Recon policy without type:")
    logger.info(pp.pformat(policy_by_id))

    assert policy_by_id is not None


@pytest.mark.integration
def test_list_all_recon_policies(adoc_client):
    filter_ = PolicyFilter(
        policyType=RuleType.RECONCILIATION,
        enable=True,
    )

    policies = adoc_client.list_all_policies(filter=filter_)

    logger.info("Recon policies list:")
    logger.info(pp.pformat(policies))

    assert policies is not None


@pytest.mark.integration
def test_execute_recon_policy_async(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )

    def operation():
        return policy_response.execute(sync=False)

    executor = retry_operation(
        operation,
        test_const.MAX_RETRIES,
        test_const.RETRY_INTERVAL,
    )

    if executor.errorMessage:
        pytest.skip(executor.errorMessage)

    status = executor.get_status()
    logger.info("Recon execution status:")
    logger.info(pp.pformat(status))

    result = executor.get_result()
    logger.info("Recon execution result:")
    logger.info(pp.pformat(result))

    assert result is not None


@pytest.mark.integration
def test_cancel_recon_policy(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )

    executor = policy_response.execute(sync=False)

    if executor.errorMessage:
        pytest.skip("Execution did not start")

    executor.cancel()
    rule_result = executor.get_result(executor.id)
    logger.info("Recon cancel response:")
    logger.info(pp.pformat(rule_result))

    assert rule_result is not None


@pytest.mark.integration
def test_get_recon_policy_executions(adoc_client):
    policy_response = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )

    executions = adoc_client.policy_executions(
        policy_response.rule.id,
        RuleType.RECONCILIATION,
    )

    logger.info("Recon executions:")
    logger.info(pp.pformat(executions))

    assert executions is not None



import logging
import time

LOGGER = logging.getLogger(__name__)


def retry_operation(operation, max_retries, retry_interval, *args, **kwargs):
    retry_count = 0
    max_attempts = max_retries + 1  # ensure at least one execution
    last_exception = None

    while retry_count < max_attempts:
        try:
            LOGGER.info("Attempt %s of %s", retry_count + 1, max_attempts)
            result = operation(*args, **kwargs)

            if result is None:
                raise ValueError("Operation returned None")

            return result

        except Exception as e:
            last_exception = e
            LOGGER.error("Attempt %s failed: %s", retry_count + 1, e)
            retry_count += 1

            if retry_count < max_attempts:
                LOGGER.info("Retrying in %s seconds...", retry_interval)
                time.sleep(retry_interval)


    raise last_exception



import pytest
import pprint
import logging
from datetime import datetime

from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
from acceldata_sdk.models.job import JobMetadata, Node, CreateJob
from acceldata_sdk.models.pipeline import (
    CreatePipeline,
    PipelineMetadata,
    PipelineRunResult,
    PipelineRunStatus,
)

from . import test_constants as test_const

# ------------------------------------------------------------------
# Logging setup
# ------------------------------------------------------------------
logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=4)

# ------------------------------------------------------------------
# Global variables
# ------------------------------------------------------------------
parent_span_context = None
pipeline_run_id = None


@pytest.fixture(scope="module", autouse=True)
def setup_pipeline_and_run(adoc_client):
    global parent_span_context, pipeline_run_id

    logger.info("===== entering create_pipeline_and_run =====")

    meta = PipelineMetadata(
        owner="sdk/pipeline-user",
        team="ADOC",
        codeLocation="...",
    )

    pipeline_name_ = test_const.PIPELINE_NAME

    pipeline = CreatePipeline(
        uid=test_const.PIPELINE_UID,
        name=pipeline_name_,
        description=(
            f"The pipeline {pipeline_name_} has been created "
            f"from acceldata-sdk using External integration"
        ),
        meta=meta,
        context={
            "pipeline_uid": test_const.PIPELINE_UID,
            "pipeline_name": pipeline_name_,
        },
    )

    pipeline_res = adoc_client.create_pipeline(pipeline=pipeline)
    logger.info("Created pipeline with pipeline id :: %s", pipeline_res.id)

    pipeline_run = pipeline_res.create_pipeline_run()
    pipeline_run_id = pipeline_run.id

    span_name_ = f"{test_const.PIPELINE_UID}.root.span"
    parent_span_context = pipeline_run.create_span(uid=span_name_)

    logger.info("Starting root span for pipeline run")
    parent_span_context.start()

    yield

    # ---------------- teardown ----------------
    logger.info("Tearing down pipeline run")

    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    pipeline_run = pipeline.get_run(pipeline_run_id)

    parent_span = pipeline_run.get_root_span()
    parent_span.end(
        context_data={"dag_status": "SUCCESS", "time": str(datetime.now())}
    )

    pipeline_run.update_pipeline_run(
        context_data={"status": "success"},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED,
    )

    logger.info("Pipeline run completed successfully")


def create_job_span_not_bounded(
        adoc_client,
        job_uid,
        inputs,
        outputs,
        metadata,
        context_job,
        span_uid,
):
    logger.info(
        "Entering create_job_span_not_bounded for job_uid=%s",
        job_uid,
    )

    span_uid_temp = span_uid

    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    pipeline_run = pipeline.get_run(pipeline_run_id)

    try:
        job = CreateJob(
            uid=job_uid,
            name=f"{job_uid} Job",
            pipeline_run_id=pipeline_run.id,
            description=f"{job_uid} created using torch job decorator",
            inputs=inputs,
            outputs=outputs,
            meta=metadata,
            context=context_job,
        )

        job = pipeline.create_job(job)
        logger.info("Create job response: %s", job)

    except Exception as e:
        logger.exception("Error while creating job %s", job_uid)
        raise

    logger.info("Successfully created job %s", job_uid)

    parent_span_context1 = pipeline_run.get_root_span()

    associated_job_uids = [job_uid]

    if span_uid is None:
        span_uid_temp = job_uid

    span_context = parent_span_context1.create_child_span(
        uid=span_uid_temp,
        context_data={"time": str(datetime.now())},
        associatedJobUids=associated_job_uids,
    )

    logger.info(
        "Created child span uid=%s for job_uid=%s",
        span_uid_temp,
        job_uid,
    )

    return span_context


# ------------------------------------------------------------------
# Tests
# ------------------------------------------------------------------
@pytest.mark.integration
@pytest.mark.order(1)
def test_write_file_func(adoc_client):
    logger.info("===== entering write_file_func =====")

    context_job = {
        "job": "data_gene",
        "uid": "customers.data-generation",
        "operator": "write_file_func",
    }

    span_context_parent = create_job_span_not_bounded(
        adoc_client=adoc_client,
        job_uid="customers.data-generation",
        inputs=[],
        outputs=[Node(job_uid="customers.s3-upload")],
        metadata=JobMetadata(
            "Jason",
            "COKE",
            "https://github.com/coke/reports/customers.kt",
        ),
        context_job=context_job,
        span_uid="customers.data.generation",
    )

    size = test_const.CSV_SIZE

    logger.info("Starting span customers.data.generation")
    span_context_parent.start()

    span_context_parent.send_event(
        GenericEvent(
            context_data={
                "Size": size - 1,
                "total_file": 1,
                "schema": "name,address,dept_id",
            },
            event_uid="customers.data.generation.metadata",
        )
    )

    logger.info("Sent GenericEvent for data generation")

    span_context_parent.send_event(
        LogEvent(
            context_data={
                "Size": size - 1,
                "total_file": 1,
                "schema": "name,address,dept_id",
            },
            log_data="Customer data generated successfully.",
        )
    )

    logger.info("Sent LogEvent for data generation")

    span_context_parent.end()
    logger.info("Ended span customers.data.generation")

@pytest.mark.integration
@pytest.mark.order(2)
def test_upload_file_to_s3(adoc_client):
    logger.info("===== entering upload_file_to_s3 =====")

    context_job = {
        "job": "data_upload",
        "time": str(datetime.now()),
        "uid": "customers.s3-upload",
        "operator": "upload_file_to_s3",
    }

    span_context_parent = create_job_span_not_bounded(
        adoc_client=adoc_client,
        job_uid="customers.s3-upload",
        inputs=[Node(job_uid="customers.data-generation")],
        outputs=[
            Node(
                asset_uid=f"{test_const.S3_DS}.{test_const.S3_CUSTOMER}"
            )
        ],
        metadata=JobMetadata(
            "Jason",
            "COKE",
            "https://github.com/coke/reports/customers.kt",
        ),
        context_job=context_job,
        span_uid="customers.s3.upload",
    )

    span_context_parent.start()
    logger.info("Started span customers.s3.upload")

    span_context_parent.end()
    logger.info("Ended span customers.s3.upload")

@pytest.mark.integration
@pytest.mark.order(3)
def test_rds_and_s3_clubbing(adoc_client):
    logger.info("===== entering rds_and_s3_clubbing =====")

    context_job = {
        "job": "data_clubbing",
        "time": str(datetime.now()),
        "uid": "customers.s3-snowflake-clubbing",
        "operator": "rds_and_s3_clubbing",
    }

    parent_span_context1 = create_job_span_not_bounded(
        adoc_client=adoc_client,
        job_uid="customers.s3-snowflake-clubbing",
        inputs=[
            Node(
                asset_uid=f"{test_const.S3_DS}.{test_const.S3_CUSTOMER}"
            ),
            Node(
                asset_uid=f"{test_const.SDK_SNOWFLAKE_DATA_SOURCE}.{test_const.SNOWFLAKE_SERVICES}"
            ),
        ],
        outputs=[
            Node(
                asset_uid=f"{test_const.SDK_SNOWFLAKE_DATA_SOURCE}.{test_const.S3_SNOWFLAKE_CUSTOMERS}"
            )
        ],
        metadata=JobMetadata(
            "BEN",
            "COKE",
            "https://github.com/coke/reports/rds_customers.kt",
        ),
        context_job=context_job,
        span_uid="customers.s3.snowflake.clubbing",
    )

    parent_span_context1.start()
    logger.info("Started span customers.s3.snowflake.clubbing")

    read_csv_span = parent_span_context1.create_child_span(uid="read_csv")
    read_csv_span.start()
    logger.info("Started read_csv span")

    read_csv_span.send_event(
        GenericEvent(
            context_data={
                "total_file_to_be_read": 1,
                "RDS_TABLES": "1",
            },
            event_uid="read_csv_data",
        )
    )

    read_csv_span.end()
    logger.info("Ended read_csv span")

    parent_span_context1.send_event(
        GenericEvent(
            context_data={
                "total_file_to_be_read": 1,
                "RDS_TABLES": "1",
                "DATA_INSERTED": 100,
                "RDS_USER": "snowflake",
            },
            event_uid="customers.rds.migration.metadata",
        )
    )

    logger.info("Sent RDS migration metadata event")

    parent_span_context1.end()
    logger.info("Ended span customers.s3.snowflake.clubbing")



import logging
import pprint
import random
from datetime import datetime

import acceldata_sdk.constants as const
import pytest
from acceldata_sdk.errors import APIError
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import (
    CreatePipeline,
    PipelineMetadata,
    PipelineRunResult,
    PipelineRunStatus,
)
from acceldata_sdk.models.profile import ProfilingType
from acceldata_sdk.models.ruleExecutionResult import RuleType, PolicyFilter
from acceldata_sdk.models.tags import AssetLabel, CustomAssetMetadata

from . import test_constants as test_const

pp = pprint.PrettyPrinter(indent=4)
logger = logging.getLogger(__name__)


# ============================================================
# TORCH VERSION
# ============================================================

@pytest.mark.integration
def test_get_torch_version(adoc_client):
    version = adoc_client.get_torch_version()
    logger.info("Torch version: %s", version)
    assert version is not None


@pytest.mark.integration
def test_get_supported_sdk_versions(adoc_client):
    versions = adoc_client.get_supported_sdk_versions()
    logger.info("Supported SDK versions: %s", versions)
    assert versions is not None


# ============================================================
# PIPELINE
# ============================================================

@pytest.mark.integration
def test_create_pipeline(adoc_client):
    meta = PipelineMetadata(
        owner="sdk/pipeline-user",
        team="TORCH",
        codeLocation="...",
    )

    pipeline = CreatePipeline(
        uid=test_const.PIPELINE_UID,
        name=test_const.PIPELINE_NAME,
        description="Created from torch-sdk",
        meta=meta,
        context={
            "pipeline_uid": test_const.PIPELINE_UID,
            "pipeline_name": test_const.PIPELINE_NAME,
        },
    )

    res = adoc_client.create_pipeline(pipeline)
    fetched = adoc_client.get_pipeline(res.id)

    assert fetched is not None


@pytest.mark.integration
def test_get_pipeline_by_uid(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    assert pipeline is not None


@pytest.mark.integration
def test_create_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.create_pipeline_run()
    fetched = pipeline.get_run(run.id)
    assert run == fetched


@pytest.mark.integration
def test_create_pipeline_run_with_continuation_id(adoc_client):
    pipeline = adoc_client.create_pipeline(
        CreatePipeline(
            uid=test_const.PIPELINE_UID_CONTINUATION,
            name=test_const.PIPELINE_UID_CONTINUATION,
        )
    )

    run = pipeline.create_pipeline_run(
        continuation_id=f"run.{random.random()}"
    )

    fetched = pipeline.get_run(run.id)
    assert fetched == run


@pytest.mark.integration
def test_get_pipelines(adoc_client):
    pipelines = adoc_client.get_pipelines()
    assert pipelines


@pytest.mark.integration
def test_latest_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    latest = pipeline.get_latest_pipeline_run()
    runs = pipeline.get_runs()
    assert latest == runs[0]


# ============================================================
# SPANS & JOBS
# ============================================================

@pytest.mark.integration
def test_create_root_span(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()
    span = run.create_span(uid=f"{test_const.PIPELINE_UID}.span")
    assert span is not None


@pytest.mark.integration
def test_create_job_pipeline(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    job = CreateJob(
        uid=test_const.JOB_UID_READ,
        name="customers read",
        pipeline_run_id=run.id,
        inputs=[Node(asset_uid="S3-DS.s3_customers")],
        outputs=[Node(job_uid=test_const.JOB_UID_GENERATE)],
        meta=JobMetadata("DR", "DR team", "https://github.com"),
        context={"time": str(datetime.now())},
        bounded_by_span=True,
        span_uid=f"{test_const.JOB_UID_READ}.span",
    )

    created = pipeline.create_job(job)
    span = run.get_span(f"{test_const.JOB_UID_READ}.span")
    span.end()

    assert created is not None


@pytest.mark.integration
def test_create_child_span(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    parent = run.get_span(f"{test_const.PIPELINE_UID}.span")

    child = parent.create_child_span(
        uid=f"{test_const.JOB_UID_SALES}.span",
        context_data={"time": str(datetime.now())},
        associatedJobUids=[test_const.JOB_UID_SALES],
    )

    child.start()
    child.send_event(
        GenericEvent(
            event_uid="customers.data.metadata",
            context_data={"files": 1},
        )
    )
    child.send_event(
        LogEvent(
            log_data="Sales aggregated",
            context_data={"rows": 100},
        )
    )
    child.end()

    assert child is not None


@pytest.mark.integration
def test_pipeline_run_details(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()
    details = run.get_details()
    assert details is not None


@pytest.mark.integration
def test_update_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.PIPELINE_UID)
    run = pipeline.get_latest_pipeline_run()

    res = run.update_pipeline_run(
        context_data={"env": "backend"},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED,
    )

    assert res is not None


@pytest.mark.integration
def test_delete_pipeline(adoc_client):
    pipeline = adoc_client.get_pipeline(
        test_const.PIPELINE_UID_CONTINUATION
    )
    res = pipeline.delete()
    assert res is not None


# ============================================================
# DQ POLICY
# ============================================================

@pytest.mark.integration
def test_get_dq_policy(adoc_client):
    policy = adoc_client.get_policy(
        const.PolicyType.DATA_QUALITY,
        test_const.DQ_POLICY_NAME,
    )
    assert policy is not None


@pytest.mark.integration
def test_list_dq_policies(adoc_client):
    rules = adoc_client.list_all_policies(
        filter=PolicyFilter(
            policyType=RuleType.DATA_QUALITY,
            enable=True,
        )
    )
    assert rules


# ============================================================
# RECON POLICY
# ============================================================

@pytest.mark.integration
def test_get_recon_policy(adoc_client):
    policy = adoc_client.get_policy(
        const.PolicyType.RECONCILIATION,
        test_const.RECON_POLICY_NAME,
    )
    assert policy is not None


# ============================================================
# DATASOURCE
# ============================================================

@pytest.mark.integration
def test_get_datasource(adoc_client):
    ds = adoc_client.get_datasource(test_const.DS_NAME, True)
    assert ds is not None


@pytest.mark.integration
def test_get_datasource_by_id(adoc_client):
    ds = adoc_client.get_datasource(test_const.DS_NAME, True)
    ds_by_id = adoc_client.get_datasource(ds.id, False)
    assert ds_by_id is not None


@pytest.mark.integration
def test_start_crawler(adoc_client):
    ds = adoc_client.get_datasource(test_const.DS_NAME, False)
    try:
        res = ds.start_crawler()
    except APIError as e:
        logger.info(e)

        msg = str(e)

        if "422" in msg and "Unable to start the Crawler" in msg:
            pytest.skip("Unable to start the Crawler")

        raise


# ============================================================
# ASSETS
# ============================================================

@pytest.mark.integration
def test_get_asset(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)
    assert asset is not None


@pytest.mark.integration
def test_asset_metadata(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)
    metadata = asset.get_metadata()
    assert metadata is not None


@pytest.mark.integration
def test_asset_labels(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)
    labels = asset.get_labels()
    assert labels is not None


@pytest.mark.integration
def test_add_asset_labels(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)
    asset.add_labels(
        [
            AssetLabel("sdk", "test"),
        ]
    )
    labels = asset.get_labels()
    assert labels is not None


@pytest.mark.integration
def test_asset_custom_metadata(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)
    asset.add_custom_metadata(
        [
            CustomAssetMetadata("sdk_key", "sdk_value"),
        ]
    )
    metadata = asset.get_metadata()
    assert metadata is not None


# ============================================================
# PROFILING
# ============================================================

@pytest.mark.integration
def test_execute_profile(adoc_client):
    asset = adoc_client.get_asset(test_const.ASSET_UID_TABLE)

    try:
        profile = asset.start_profile(ProfilingType.FULL)
        status = profile.get_status()
        assert status is not None

    except APIError as e:
        logger.info(e)

        msg = str(e)

        if "422" in msg and "Asset cannot be profiled" in msg:
            pytest.skip("Asset not profileable in current environment")

        raise



import logging

import pytest
from acceldata_sdk.errors import APIError
from acceldata_sdk.errors import TorchSdkException
from acceldata_sdk.models.common_types import (
    PolicyExecutionRequest,
    ExecutionType,
    AssetMarkerConfig,
    BoundsDateTimeMarkerConfig,
    BoundsFileEventMarkerConfig,
    TimestampBasedMarkerConfig,
)

from . import test_constants as test_const
from ..commons.retry import retry_operation

logger = logging.getLogger(__name__)


# ============================================================
# BACKWARD COMPATIBLE RECON EXECUTION
# ============================================================
def execute_policy_or_accept_running(operation):
    try:
        return retry_operation(
            operation,
            test_const.MAX_RETRIES,
            test_const.RETRY_INTERVAL,
        )

    except APIError as e:
        msg = str(e)
        logger.info(msg)

        # ✅ ACCEPTABLE / PASS CONDITIONS
        if (
                "Previous execution of rule" in msg
                and "has not completed" in msg
        ):
            # Treat as success
            logger.info("Previous execution still running – accepting as PASS")
            return {"status": "ALREADY_RUNNING"}

        # ❌ Anything else is a real failure
        raise


@pytest.mark.integration
def test_execute_full_recon_backward_compatible(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.SELECTIVE_RECON_POLICY_BACKWARD_COMPATIBLE_NAME
    )

    logger.info("Executing FULL reconciliation (backward compatible)")
    logger.info("Policy id: %s", policy.id)

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            incremental=False,
        )

    result = execute_policy_or_accept_running(
        operation)
    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_incremental_recon_backward_compatible(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_RECON_POLICY_NAME
    )

    logger.info("Executing INCREMENTAL reconciliation (backward compatible)")
    logger.info("Policy id: %s", policy.id)

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            incremental=True,
        )

    result = execute_policy_or_accept_running(
        operation)

    logger.info("Execution result: %s", result)
    assert result is not None


# ============================================================
# POLICY EXECUTION REQUEST — FULL / INCREMENTAL
# ============================================================

@pytest.mark.integration
def test_execute_full_recon_using_execution_request(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.SELECTIVE_RECON_POLICY_BACKWARD_COMPATIBLE_NAME
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.FULL
    )

    logger.info("Executing FULL reconciliation using PolicyExecutionRequest")

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(
        operation)

    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_incremental_recon_using_execution_request(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_RECON_POLICY_NAME
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.INCREMENTAL
    )

    logger.info("Executing INCREMENTAL reconciliation using PolicyExecutionRequest")

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(
    operation)

    logger.info("Execution result: %s", result)
    assert result is not None


# ============================================================
# SELECTIVE RECON — VALIDATION
# ============================================================

@pytest.mark.integration
def test_execute_selective_recon_without_marker_config(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_DQ_ID
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE
    )

    logger.info("Executing SELECTIVE reconciliation without marker config")

    with pytest.raises(TorchSdkException):
        adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )


# ============================================================
# SELECTIVE RECON — DATETIME BASED
# ============================================================

@pytest.mark.integration
def test_execute_selective_recon_datetime_based(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_RECON_POLICY_NAME
    )

    marker = BoundsDateTimeMarkerConfig(
        dateColumnName="TO_DATE",
        format="yyyy-MM-dd",
        fromDate="2023-07-01 00:00:00.000",
        toDate="2024-07-14 23:59:59.999",
        timeZoneId="Asia/Calcutta",
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.INCREMENTAL_DQ_RECON_TIME_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    logger.info("Executing SELECTIVE reconciliation (datetime based)")

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(
        operation)

    logger.info("Execution result: %s", result)
    assert result is not None

# ============================================================
# SELECTIVE RECON — FILE EVENT BASED
# ============================================================

# @pytest.mark.integration
# def test_execute_selective_recon_file_event_based(adoc_client):
#     policy = adoc_client.get_policy(
#         identifier=test_const.FILE_EVENT_BASED_RECON_POLICY
#     )
#
#     marker = BoundsFileEventMarkerConfig(
#         fromDate="2020-07-01 00:00:00.000",
#         toDate="2024-07-01 23:59:59.999",
#         timeZoneId="Asia/Calcutta",
#     )
#
#     request = PolicyExecutionRequest(
#         executionType=ExecutionType.SELECTIVE,
#         markerConfigs=[
#             AssetMarkerConfig(
#                 assetId=test_const.ASSET_ID,
#                 markerConfig=marker,
#             )
#         ],
#     )
#
#     logger.info("Executing SELECTIVE reconciliation (file event based)")
#
#     def operation():
#         return adoc_client.execute_reconciliation_rule(
#             rule_id=policy.id,
#             policy_execution_request=request,
#         )
#
#     result = retry_operation(
#         operation,
#         test_const.MAX_RETRIES,
#         test_const.RETRY_INTERVAL,
#     )
#
#     logger.info("Execution result: %s", result)
#     assert result is not None
#

# ============================================================
# SELECTIVE RECON — KAFKA TIMESTAMP BASED
# ============================================================

@pytest.mark.integration
def test_execute_selective_recon_kafka_timestamp_based(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.KAFKA_RECON_POLICY_NAME
    )

    marker = TimestampBasedMarkerConfig(
        format="yyyy-mm-dd",
        initialOffset="2023-06-01",
        timeZoneId="Asia/Calcutta",
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.KAFKA_DQ_POLICY_BACKING_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    logger.info("Executing SELECTIVE reconciliation (kafka timestamp based)")

    def operation():
        return adoc_client.execute_reconciliation_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(
        operation)

    logger.info("Execution result: %s", result)
    assert result is not None



import logging

import pytest
from acceldata_sdk.errors import TorchSdkException, APIError
from acceldata_sdk.models.common_types import (
    PolicyExecutionRequest,
    ExecutionType,
    AssetMarkerConfig,
    BoundsIdMarkerConfig,
    BoundsDateTimeMarkerConfig,
    BoundsFileEventMarkerConfig,
    TimestampBasedMarkerConfig,
    YunikornConfig,
    SparkResourceConfig,
    RuleSparkSQLDynamicFilterVariableMapping,
    Mapping,
)

from . import test_constants as test_const
from ..commons.retry import retry_operation

logger = logging.getLogger(__name__)


def execute_policy_or_accept_running(operation):
    try:
        return retry_operation(
            operation,
            test_const.MAX_RETRIES,
            test_const.RETRY_INTERVAL,
        )

    except APIError as e:
        msg = str(e)
        logger.info(msg)

        # ✅ ACCEPTABLE / PASS CONDITIONS
        if (
                "Previous execution of rule" in msg
                and "has not completed" in msg
        ):
            # Treat as success
            logger.info("Previous execution still running – accepting as PASS")
            return {"status": "ALREADY_RUNNING"}

        # ❌ Anything else is a real failure
        raise


@pytest.mark.integration
def test_execute_full_dq_backward_compatible(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.DQ_POLICY_BACKWARD_COMPATIBLE_NAME
    )

    logger.info("Executing FULL DQ (backward compatible)")

    def operation():
        return adoc_client.execute_dq_rule(rule_id=policy.id)

    result = execute_policy_or_accept_running(operation)
    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_incremental_dq_backward_compatible(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_DQ_POLICY_NAME
    )

    logger.info("Executing INCREMENTAL DQ (backward compatible)")

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            incremental=True,
        )

    result = execute_policy_or_accept_running(operation)
    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_full_dq_using_execution_request(adoc_client):
    policy = adoc_client.get_policy(identifier=test_const.DQ_POLICY_NAME)

    request = PolicyExecutionRequest(
        executionType=ExecutionType.FULL
    )

    logger.info("Executing FULL DQ using PolicyExecutionRequest")

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_incremental_dq_using_execution_request(adoc_client):
    policy = adoc_client.get_policy(identifier=test_const.INCREMENTAL_DQ_POLICY_NAME)

    request = PolicyExecutionRequest(
        executionType=ExecutionType.INCREMENTAL
    )

    logger.info("Executing INCREMENTAL DQ using PolicyExecutionRequest")

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    logger.info("Execution result: %s", result)
    assert result is not None


@pytest.mark.integration
def test_execute_selective_dq_id_based(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_DQ_ID
    )

    marker = BoundsIdMarkerConfig(
        idColumnName="ID",
        fromId=0,
        toId=1000,
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.INCREMENTAL_DQ_DATE_TIME_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    logger.info("Executing SELECTIVE DQ (ID based)")

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    assert result is not None


@pytest.mark.integration
def test_execute_selective_dq_without_marker_config(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.INCREMENTAL_DQ_ID
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE
    )

    logger.info("Executing SELECTIVE DQ without marker config")

    with pytest.raises(TorchSdkException):
        adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )


@pytest.mark.integration
def test_execute_selective_dq_datetime_based(adoc_client):
    policy = adoc_client.get_policy(identifier=test_const.INCREMENTAL_DQ_DATE_TIME)

    marker = BoundsDateTimeMarkerConfig(
        dateColumnName="CURRENT_DATE",
        format="yyyy-MM-dd",
        fromDate="2023-07-01 00:00:00.000",
        toDate="2026-01-01 23:59:59.999",
        timeZoneId="Asia/Calcutta",
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.INCREMENTAL_DQ_DATE_TIME_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    assert result is not None


@pytest.mark.integration
def test_execute_selective_dq_file_event_based(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.FILE_EVENT_BASED_DQ_POLICY
    )

    marker = BoundsFileEventMarkerConfig(
        fromDate="2026-01-01 00:00:00.000",
        toDate="2026-02-06 23:59:59.999",
        timeZoneId="Asia/Calcutta",
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.FILE_EVENT_BASED_DQ_BACKING_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    assert result is not None


@pytest.mark.integration
def test_execute_selective_dq_kafka_timestamp_based(adoc_client):
    policy = adoc_client.get_policy(
        identifier=test_const.KAFKA_DQ_POLICY_NAME
    )

    marker = TimestampBasedMarkerConfig(
        format="yyyy-mm-dd",
        initialOffset="2023-06-01",
        timeZoneId="Asia/Calcutta",
    )

    request = PolicyExecutionRequest(
        executionType=ExecutionType.SELECTIVE,
        markerConfigs=[
            AssetMarkerConfig(
                assetId=test_const.KAFKA_DQ_POLICY_BACKING_ASSET_ID,
                markerConfig=marker,
            )
        ],
    )

    def operation():
        return adoc_client.execute_dq_rule(
            rule_id=policy.id,
            policy_execution_request=request,
        )

    result = execute_policy_or_accept_running(operation)

    assert result is not None



from dataclasses import dataclass, asdict
from enum import Enum
from typing import List, Optional, Dict

from acceldata_sdk.models.job import CreateJob
from acceldata_sdk.models.span import Span
from acceldata_sdk.models.span_context import SpanContext
from acceldata_sdk.utils.datetime_utils import DateTimeFormatter
from datetime import datetime
import logging

logger = logging.getLogger('pipeline')
logger.setLevel(logging.INFO)


@dataclass
class Tag:

    def __init__(self, id, name, displayName, *args, **kwargs):
        self.id = id
        self.name = name
        self.displayName = displayName
    id = None
    name = None
    displayName = None
    createdAt = None
    updatedAt = None


@dataclass
class PipelineMetadata:
    """
    Description:
        Pipeline metadata.

    :param owner : (String) owner of the pipeline
    :param team: (String) team of the owner
    :param codeLocation: (String) location of the code
    """
    owner: str = None
    team: str = None
    codeLocation: str = None

class PipelineSourceType(Enum):
    AIRFLOW='AIRFLOW',
    AZURE_DATA_FACTORY='AZURE_DATA_FACTORY',
    OTHERS='OTHERS'

    def to_dict(self):
        return self.name

class CreatePipeline:
    # createdAt = None
    # enabled = None
    # id = None
    # interrupt = None
    # notificationChannels = None
    # schedule = None
    # scheduled = None
    # schedulerType = None
    # updatedAt = None

    def __init__(self, uid: str, name: str,
                 description: str = None, meta: PipelineMetadata = None, createdAt: datetime = None,
                 updatedAt=None, sourceType=None, **context):
        """
        Description:
            Class used for create pipeline in torch catalog
        :param uid: uid of the pipeline
        :param name: name of the pipeline
        :param description: (Optional)description of the pipeline
        :param meta: (PipelineMetadata) meta data of the pipeline
        :param createdAt: (Optional) creation time of the pipeline. If not set, the current time will be used
        :param updatedAt: (Optional) update time of the pipeline. If not set, the current time will be used
        :param context: context map of the pipeline
        """
        self.uid = uid
        self.name = name
        self.description = description
        self.context = context
        if meta is not None:
            self.meta = PipelineMetadata(owner=meta.owner, team=meta.team, codeLocation=meta.codeLocation)
        else:
            self.meta = None
        self.createdAt = DateTimeFormatter.format_datetime(createdAt)
        self.updatedAt = DateTimeFormatter.format_datetime(updatedAt)
        if sourceType is not None:
            self.sourceType = sourceType
        else:
            self.sourceType = None


    def __eq__(self, other):
        return self.uid == other.uid

    def __repr__(self):
        return f"Pipeline({self.uid!r})"

    def add_context(self, key: str, value: str):
        """
        Used to add context in context map
        :param key: key of the context
        :param value: value of the context
        :return: context dir
        """
        self.context[key] = value
        return self.context

class PipelineRunStatus(Enum):
    """
        pipeline run status
    """
    STARTED = 1
    COMPLETED = 2
    FAILED = 3
    ABORTED = 4


class PipelineRunResult(Enum):
    """
        Pipeline run result
    """
    RUNNING = 'RUNNING'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'
    CANCELLED = 'CANCELLED'


class NodeType(Enum):
    """
        Pipeline Node Type
    """
    ASSET = 'ASSET'
    FUNCTIONAL = 'FUNCTIONAL'


class NodeAddition(Enum):
    """
        Pipeline Node Addition
    """
    PIPELINE_CREATION = 'PIPELINE_CREATION'
    MANUAL_ADDITION = 'MANUAL_ADDITION'
    QUERY_LOG_ANALYSIS = 'QUERY_LOG_ANALYSIS'


class NodeStatus(Enum):
    """
        Pipeline Node Status
    """
    ACTIVE = 'ACTIVE'
    STALE = 'STALE'
    DELETED = 'DELETED'


class PipelineRun:

    def __init__(self, pipelineId, versionId=None, status=PipelineRunStatus.STARTED, id=None, startedAt: datetime = None,
                 finishedAt: datetime = None, result=None, client=None, args=None,
                 continuationId=None, successEvents=None, errorEvents=None, warningEvents=None, avgExecutionTime=None,
                 *argumentss, **kwargs):
        """
        Description:
                Pipeline run used for instantiate run of the pipeline. To create new pipeline run you need to pass pipelineId, version.
        :param pipelineId: pipleline id
        :param versionId:  pipeline run current version
        :param status: status of the pipeline run
        :param id: pipeline run id
        :param startedAt: An optional parameter used to define the starting time of the pipeline run. If this parameter is not provided, the pipeline will commence at the time of its creation by default.
        :param finishedAt: An optional parameter used to specify the completion time of the pipeline run. If this parameter is not provided, it will be automatically set to the time when the pipeline run is finished.
        :param args: additional args for pipeline run
        :param result: pipeline run result
        """
        self.pipelineId = pipelineId
        self.versionId = versionId
        self.status = status
        self.args = args
        self.result = result
        if continuationId is not None:
            self.continuationId = continuationId
        if successEvents is not None:
            self.successEvents = successEvents
        if errorEvents is not None:
            self.errorEvents = errorEvents
        if warningEvents is not None:
            self.warningEvents = warningEvents
        if avgExecutionTime is not None:
            self.avgExecutionTime = avgExecutionTime
        self.startedAt = startedAt
        self.finishedAt = finishedAt

        if id is not None:
            self.id = id
            self.client = client

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return f"PipelineRun({self.__dict__})"

    # convert job object to dict type
    def _convert_job_to_dict(self, job: CreateJob):
        """
        Description:
            Convert createJob class instance to dict type
        :param job: createJob class instance
        :return: dict form of createJob class instance
        """
        job_dict = job.__dict__
        if hasattr(job, 'meta'):
            meta = asdict(job.meta)
            job_dict['meta'] = meta
        input_dict = []
        for ds_in in job.inputs:
            input_dict.append(asdict(ds_in, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}))
        output_dict = []
        for ds_out in job.outputs:
            output_dict.append(asdict(ds_out, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}))
        job_dict['inputs'] = input_dict
        job_dict['outputs'] = output_dict
        job_dict['version'] = self.versionId
        return job_dict

    # create job for pipeline run
    def create_job(self, job: CreateJob):
        """
        Description:
            used to create job for any pipeline run
        :return: Job class instance of created job        """
        if job.uid is None or job.name is None:
            raise Exception('To create a job job uid and name are required')
        payload = self._convert_job_to_dict(job)
        res = self.client.create_job(payload, self.pipelineId)
        if job.bounded_by_span:
            if job.span_uid is not None:
                associated_job_uids = [job.uid]
                span_uid_temp = job.span_uid
                # Get root span to create child span under
                parent_span_context = self.get_root_span()
                if parent_span_context is not None:
                    # Create child span under root span
                    span_context = parent_span_context.create_child_span(
                        uid=span_uid_temp,
                        associatedJobUids=associated_job_uids,
                        with_explicit_time=job.with_explicit_time
                    )
                else:
                    # Create root span under pipeline_run as root span doesn't exist
                    self.create_span(
                        uid=span_uid_temp,
                        associatedJobUids=associated_job_uids,
                        with_explicit_time=job.with_explicit_time
                    )
            else:
                raise Exception('To create a span span_uid is required')
        return res

    # convert pipeline run to dict type
    def _convert_pipeline_run_to_dict(self, pipeline_run=None):
        """
            Description:
                Convert PipelineRun class instance to dict type
            :param pipeline_run: PipelineRun class instance
            :return: dict form of PipelineRun class instance
        """
        payload = pipeline_run.__dict__
        payload['status'] = pipeline_run.status.name
        if pipeline_run.result is not None:
            payload['result'] = pipeline_run.result.name
        else:
            del payload['result']
        pipeline_run_payload = {'run': payload}
        return pipeline_run_payload

    # update run for a pipeline
    def update_pipeline_run(self, result: PipelineRunResult,
                            status: PipelineRunStatus = PipelineRunStatus.STARTED,
                            context_data: {} = None, finishedAt: datetime = None):
        """
        Description:
            used to update an existing pipeline run
        :return: updated pipelineRun class instance
        """
        if finishedAt is not None:
            finishedAt = finishedAt.isoformat()

        update_pipeline_run = PipelineRun(
            pipelineId=self.pipelineId,
            versionId=self.versionId,
            result=result,
            args=context_data,
            status=status,
            finishedAt = finishedAt
        )
        payload = self._convert_pipeline_run_to_dict(update_pipeline_run)
        res = self.client.update_pipeline_run(self.id, payload)
        return res

    # convert span to dict
    def _convert_span_to_dict(self, span: Span, associatedJobUids):
        """
            Description:
                Convert Span class instance to dict type
            :param span: Span class instance
            :return: dict form of Span class instance
        """
        payload = span.__dict__
        payload['status'] = span.status.name
        span_payload = {'span': payload, 'associatedJobUids' : associatedJobUids }
        return span_payload

    # create span for pipeline run
    def create_span(self, uid: str, context_data: dict = None, associatedJobUids = None, createdAt: datetime = None, with_explicit_time: bool = False):
        """
        Description:
            used to create span for any pipeline run
        :param associatedJobUids: list of string (job uids)
        :param context_data
        :param uid: span uid
        :param with_explicit_time: A boolean flag that determines how the span should be started.
        - If set to True, you can specify the start time for the span in subsequent span events.
        - If set to False, the span will be automatically started with the current time at the moment of creation.
        :return: SpanContext of the span
        """
        if uid is None:
            raise Exception('Span uid can not be None. You need to pass span uid to create span')
        create_span = Span(
            uid=uid,
            pipelineRunId=self.id,
            createdAt=createdAt
        )
        if associatedJobUids is None:
            associatedJobUids = []
        elif not isinstance(associatedJobUids, list):
            raise Exception('associatedJobUids should be a list')

        payload = self._convert_span_to_dict(create_span, associatedJobUids)
        logger.debug("create_span method invoked with_explicit_time: ", with_explicit_time)
        res = self.client.create_span(self.id, payload)
        span_context = SpanContext(client=self.client, span=res, context_data=context_data, new_span=True,
                                   with_explicit_time=with_explicit_time)
        return span_context

    def get_span(self, span_uid: str):
        """
            Description:
                Get span of the pipeline run by uid
        :param span_uid: uid of the span
        :return: SpanContext instance of the input span uid
        """
        span = self.client.get_span(pipeline_run_id= self.id, uid=span_uid)
        span_context = SpanContext(client=self.client, span= span)
        return span_context

    def get_root_span(self):
        """
            Description:
                Get root span of the pipeline run
        :return: SpanContext instance of the root span
        """
        span_context = None
        span = self.client.get_root_span( pipeline_run_id=self.id)
        if span is not None:
            span_context = SpanContext(client=self.client, span=span)
        return span_context

    def get_details(self):
        """
            Description:
                Get root span of the pipeline run
        :return: SpanContext instance of the root span
        """
        response = self.client.get_run_details(pipeline_id=self.pipelineId, version_id=self.versionId)
        return response

    def get_spans(self):
        """
            Description:
                Get span of the pipeline run
        :return: SpanContext instance of the input span uid
        """
        spans = self.client.get_spans(pipeline_run_id=self.id)
        span_context_list = []
        for res in spans:
            span_context_list.append(SpanContext(client=self.client, span=res))
        return span_context_list


@dataclass
class Node:
    id = None
    type: NodeType = None
    assetId = None
    pipelineId = None
    name = None
    description = None
    context = None
    meta: PipelineMetadata = None
    versionId = None
    uid = None
    createdVia: NodeAddition = NodeAddition.PIPELINE_CREATION
    status: NodeStatus = NodeStatus.ACTIVE
    parentId = None

    def __init__(self,
                 id = None,
                 type: NodeType = None,
                 assetId = None,
                 pipelineId = None,
                 name = None,
                 description = None,
                 context = None,
                 meta: PipelineMetadata = None,
                 versionId = None,
                 uid = None,
                 createdVia: NodeAddition = NodeAddition.PIPELINE_CREATION,
                 status: NodeStatus = NodeStatus.ACTIVE,
                 parentId = None, *args, **kwargs
                 ):
        self.id = id
        self.type = type
        self.assetId = assetId
        self.pipelineId = pipelineId
        self.name = name
        self.description = description
        self.context = context
        if isinstance(meta, dict):
            self.meta = PipelineMetadata(**meta)
        else:
            self.meta = meta
        self.versionId = versionId
        self.uid = uid
        self.createdVia = createdVia
        self.status = status
        self.parentId = parentId

    def __repr__(self):
        return f"Node({self.__dict__})"


@dataclass 
class Edge:
    id = None
    fromId = None
    toId = None
    fromAssetType: NodeType = None
    toAssetType: NodeType = None
    versionId = None

    def __init__(self,
                 id = None,
                 fromId = None,
                 toId = None,
                 fromAssetType: NodeType = None,
                 toAssetType: NodeType = None,
                 versionId = None, *args, **kwargs
                 ):
        self.id = id
        self.fromId = fromId
        self.toId = toId
        self.fromAssetType = fromAssetType
        self.toAssetType = toAssetType
        self.versionId = versionId

    def __repr__(self):
        return f"Edge({self.__dict__})"


class Pipeline:
    def __init__(self, uid: str,
                 name: str,
                 description: str = None,
                 meta: PipelineMetadata = None,
                 createdAt: str = None,
                 enabled: bool = None,
                 id: int = None,
                 interrupt: bool = None,
                 notificationChannels=None,
                 schedule: str = None,
                 scheduled: bool = None,
                 schedulerType: str = None,
                 updatedAt=None,
                 context=None,
                 client=None,
                 tags=None,
                 *args, **kwargs):
        """
        Description:
            Class of the pipeline
        :param uid: uid of the pipeline
        :param name: name of the pipeline
        :param description: pipeline desc
        :param meta: (PipelineMetadata)metadata of the pipeline
        :param createdAt: creation time of pipeline
        :param enabled: True if pipeline is interrupted else false
        :param id: pipeline id
        :param updatedAt: updated time of the given pipeline
        :param context: context data for the pipeline (dir)
        :param kwrgs:
        """
        self.uid = uid
        self.name = name
        self.createdAt = createdAt
        self.enabled = enabled
        self.interrupt = interrupt
        self.notificationChannels = notificationChannels
        self.schedule = schedule
        self.scheduled = scheduled
        self.schedulerType = schedulerType
        self.id = id
        self.updatedAt = updatedAt
        self.description = description
        self.context = context
        if isinstance(meta, dict):
            self.meta = PipelineMetadata(**meta)
        else:
            self.meta = meta

        self.tags = list()
        if tags is not None:
            for obj in tags:
                if isinstance(obj, dict):
                    self.tags.append(Tag(**obj))
                else:
                    self.tags.append(obj)
        self.client = client

    def __repr__(self):
        return f"Pipeline({self.__dict__})"

    def __eq__(self, other):
        return self.uid == other.uid

    # convert job object to dict type
    def _convert_job_to_dict(self, job: CreateJob):
        """
        Description:
            Convert createJob class instance to dict type
        :param job: createJob class instance
        :return: dict form of createJob class instance
        """
        job_dict = job.__dict__
        if hasattr(job, 'meta'):
            meta = asdict(job.meta)
            job_dict['meta'] = meta
        input_dict = []
        for ds_in in job.inputs:
            input_dict.append(asdict(ds_in, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}))
        output_dict = []
        for ds_out in job.outputs:
            output_dict.append(asdict(ds_out, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}))
        job_dict['inputs'] = input_dict
        job_dict['outputs'] = output_dict
        pipeline_run = self.client.get_pipeline_run(job.pipeline_run_id)
        job_dict['version'] = pipeline_run.versionId
        return job_dict

    # to create job for any given pipeline
    def create_job(self, job: CreateJob):
        """
        Description:
            Used to create job and span in a pipeline
        :param job: createJob class instance that you want to add in pipeline
        :return: Job class instance of created job
        """
        if job.uid is None or job.name is None or job.pipeline_run_id is None:
            raise Exception('To create a job job uid, name and pipeline_run_id is required')

        payload = self._convert_job_to_dict(job)
        res = self.client.create_job(payload, self.id)
        if job.bounded_by_span:
            if job.pipeline_run_id is not None and job.span_uid is not None:
                associated_job_uids = [job.uid]
                span_uid_temp = job.span_uid
                # Get root span to create child span under
                pipeline_run = self.get_run(job.pipeline_run_id)
                parent_span_context = pipeline_run.get_root_span()
                if parent_span_context is not None:
                    # Create child span under root span
                    span_context = parent_span_context.create_child_span(
                        uid=span_uid_temp,
                        associatedJobUids=associated_job_uids,
                        with_explicit_time=job.with_explicit_time
                    )
                else:
                    # Create root span under pipeline_run as root span doesn't exist
                    pipeline_run.create_span(
                        uid=span_uid_temp,
                        associatedJobUids=associated_job_uids,
                        with_explicit_time=job.with_explicit_time
                    )
            else:
                raise Exception('To create a span span_uid and pipeline_run_id are required')
        return res

    # to delete pipeline
    def delete(self):
        """
        Description:
            Used to delete a pipeline
        """
        res = self.client.delete_pipeline(self.id)
        return res

    # convert pipeline run to dict type
    def _convert_pipeline_run_to_dict(self, pipeline_run: PipelineRun):
        """
            Description:
                Convert PipelineRun class instance to dict type
            :param pipeline_run: PipelineRun class instance
            :return: dict form of PipelineRun class instance
        """
        payload = pipeline_run.__dict__
        payload['status'] = pipeline_run.status.name
        if pipeline_run.result is not None:
            payload['result'] = pipeline_run.result.name
        else:
            del payload['result']
        pipeline_run_payload = {'run': payload}
        return pipeline_run_payload

    # create run for a pipeline
    def create_pipeline_run(self, context_data: {} = None, continuation_id: str = None,
                            startedAt: datetime = None) -> PipelineRun:
        """
        Description:
            used to create a pipeline run
        :param context_data: pipeline run argument
        :param continuation_id: continuation_id of pipeline run. Default value is None
        :return: pipelineRun class instance
        """
        if startedAt is not None:
            startedAt = DateTimeFormatter.format_datetime(startedAt)
            create_pipeline_run = PipelineRun(
                pipelineId=self.id,
                args=context_data,
                continuationId=continuation_id,
                startedAt=startedAt
            )
        else:
            create_pipeline_run = PipelineRun(
                pipelineId=self.id,
                args=context_data,
                continuationId=continuation_id
            )
        payload = self._convert_pipeline_run_to_dict(create_pipeline_run)
        res = self.client.create_pipeline_run(payload)
        return res

    def get_latest_pipeline_run(self) -> PipelineRun:
        return self.client.get_latest_pipeline_run(pipeline_id=self.id)

    def get_run(self, pipeline_run_id=None, continuation_id=None) -> PipelineRun:
        return self.client.get_pipeline_run(
            pipeline_run_id=pipeline_run_id,
            continuation_id=continuation_id,
            pipeline_id=self.id
        )

    def get_runs(self) -> List[PipelineRun]:
        return self.client.get_pipeline_runs(pipeline_id=self.id)


from typing import List, Optional, Dict, Any


class PipelineListingInfo:
    def __init__(
            self,
            assetNodesCount: int,
            functionalNodesCount: int,
            latestRunFinishedAt: Optional[str],
            latestRunId: Optional[int],
            latestRunResult: Optional[str],
            latestRunStartedAt: Optional[str],
            latestRunVersionId: Optional[int],
            pipelineSummary: 'PipelineSummary',
            successfulPoliciesCount: int,
            totalPoliciesCount: int,
            totalRunsCount: int,
            *args, **kwargs
    ):
        self.assetNodesCount = assetNodesCount
        self.functionalNodesCount = functionalNodesCount
        self.latestRunFinishedAt = latestRunFinishedAt
        self.latestRunId = latestRunId
        self.latestRunResult = latestRunResult
        self.latestRunStartedAt = latestRunStartedAt
        self.latestRunVersionId = latestRunVersionId
        self.pipelineSummary = pipelineSummary
        self.successfulPoliciesCount = successfulPoliciesCount
        self.totalPoliciesCount = totalPoliciesCount
        self.totalRunsCount = totalRunsCount

    def __repr__(self):
        return (
            f"PipelineListingInfo("
            f"pipelineSummary={self.pipelineSummary.name}, "
            f"assetNodesCount={self.assetNodesCount}, "
            f"functionalNodesCount={self.functionalNodesCount}, "
            f"latestRunResult={self.latestRunResult}, "
            f"totalRunsCount={self.totalRunsCount})"
        )


class PipelineSummary:
    def __init__(self, id: int, name: str, meta: 'PipelineSummaryMeta', *args, **kwargs):
        self.id = id
        self.name = name
        self.meta = meta

    def __repr__(self):
        return f"PipelineSummary(id={self.id}, name={self.name}, meta={self.meta})"


class PipelineSummaryMeta:
    def __init__(self, codeLocation: str, owner: str, team: str):
        self.codeLocation = codeLocation
        self.owner = owner
        self.team = team

    def __repr__(self):
        return (
            f"PipelineSummaryMeta(codeLocation={self.codeLocation}, "
            f"owner={self.owner}, team={self.team})"
        )


class PipelineDetails:
    pipeline: Pipeline = None
    nodes: List[Node] = list()
    edges: List[Edge] = list()

    def __init__(self, pipeline=None, nodes=None, edges=None, *args, **kwargs):
        if isinstance(pipeline, dict):
            self.pipeline = Pipeline(**pipeline)
        else:
            self.pipeline = pipeline
        self.nodes = list()
        if nodes is not None:
            for obj in nodes:
                if isinstance(obj, dict):
                    self.nodes.append(Node(**obj))
                else:
                    self.nodes.append(obj)
        self.edges = list()
        if edges is not None:
            for obj in edges:
                if isinstance(obj, dict):
                    self.edges.append(Edge(**obj))
                else:
                    self.edges.append(obj)

    def __repr__(self):
        return f"PipelineDetails({self.__dict__})"



import pytest
import logging
from datetime import datetime, timedelta

from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.events.log_events import LogEvent
from acceldata_sdk.models.job import JobMetadata, Node, CreateJob
from acceldata_sdk.models.pipeline import (
    CreatePipeline,
    PipelineMetadata,
    PipelineRunResult,
    PipelineRunStatus,
)

from . import test_constants as test_const

logger = logging.getLogger(__name__)

# =====================================================
# Runtime globals
# =====================================================
parent_span_context = None
pipeline_run_id = None

# =====================================================
# Explicit timestamps
# =====================================================
now = datetime.now()

explicit_pipeline_createdAt = now - timedelta(days=2, hours=6, minutes=30)
explicit_pipeline_updatedAt = explicit_pipeline_createdAt + timedelta(minutes=30)
explicit_pipeline_run_startedAt = explicit_pipeline_createdAt + timedelta(minutes=30)

root_span_created_at = explicit_pipeline_run_startedAt + timedelta(minutes=1)

child_span_1_created_at = explicit_pipeline_run_startedAt + timedelta(minutes=30)
child_span_1_finished_at = child_span_1_created_at + timedelta(minutes=30)

child_span_2_created_at = child_span_1_finished_at + timedelta(minutes=1)
child_span_2_finished_at = child_span_2_created_at + timedelta(minutes=10)

child_span_3_created_at = child_span_2_finished_at + timedelta(minutes=1)

child_span_3_child_1_created_at = child_span_3_created_at + timedelta(minutes=1)
child_span_3_child_1_finished_at = child_span_3_child_1_created_at + timedelta(minutes=10)

child_span_3_child_2_created_at = child_span_3_child_1_finished_at + timedelta(minutes=1)

child_span_3_child_2_child_1_created_at = child_span_3_child_2_created_at + timedelta(minutes=1)
child_span_3_child_2_child_1_finished_at = child_span_3_child_2_child_1_created_at + timedelta(minutes=10)

child_span_3_child_2_child_2_created_at = child_span_3_child_2_child_1_finished_at + timedelta(minutes=1)
child_span_3_child_2_child_2_finished_at = child_span_3_child_2_child_2_created_at + timedelta(minutes=10)

child_span_3_child_2_finished_at = child_span_3_child_2_child_2_finished_at

child_span_3_child_3_created_at = child_span_3_child_2_finished_at + timedelta(minutes=1)
child_span_3_child_3_finished_at = child_span_3_child_3_created_at + timedelta(minutes=10)

child_span_3_finished_at = child_span_3_child_3_finished_at
explicit_pipeline_run_finishedAt = child_span_3_finished_at + timedelta(minutes=1)


# =====================================================
# Pipeline + run setup
# =====================================================
@pytest.mark.integration
@pytest.mark.order(1)
def test_create_pipeline_and_run(adoc_client):
    global parent_span_context, pipeline_run_id

    logger.info("Creating pipeline with explicit timestamps")

    meta = PipelineMetadata(
        owner="sdk/pipeline-user",
        team="ADOC",
        codeLocation="...",
    )

    pipeline = CreatePipeline(
        uid=test_const.EXPLICIT_PIPELINE_UID,
        name=test_const.EXPLICIT_PIPELINE_NAME,
        description="Pipeline created using SDK explicit time flow",
        meta=meta,
        context={
            "pipeline_uid": test_const.EXPLICIT_PIPELINE_UID,
            "pipeline_name": test_const.EXPLICIT_PIPELINE_NAME,
        },
        createdAt=explicit_pipeline_createdAt,
        updatedAt=explicit_pipeline_updatedAt,
    )

    pipeline_res = adoc_client.create_pipeline(pipeline=pipeline)
    logger.info("Pipeline created id=%s", pipeline_res.id)

    pipeline_run = pipeline_res.create_pipeline_run(
        startedAt=explicit_pipeline_run_startedAt
    )

    pipeline_run_id = pipeline_run.id
    logger.info("Pipeline run started id=%s", pipeline_run_id)

    span_uid = f"{test_const.EXPLICIT_PIPELINE_UID}.root.span"

    parent_span_context = pipeline_run.create_span(
        uid=span_uid,
        with_explicit_time=True,
    )

    parent_span_context.start(created_at=root_span_created_at)
    logger.info("Root span started")


# =====================================================
# Helper
# =====================================================
def create_job_span_not_bounded(
        adoc_client,
        job_uid,
        inputs,
        outputs,
        metadata,
        context_job,
        span_uid,
):
    logger.info("Creating job %s", job_uid)

    pipeline = adoc_client.get_pipeline(test_const.EXPLICIT_PIPELINE_UID)
    pipeline_run = pipeline.get_run(pipeline_run_id)

    job = CreateJob(
        uid=job_uid,
        name=f"{job_uid} Job",
        pipeline_run_id=pipeline_run.id,
        description=f"{job_uid} created via SDK",
        inputs=inputs,
        outputs=outputs,
        meta=metadata,
        context=context_job,
        with_explicit_time=True,
    )

    job = pipeline.create_job(job)
    logger.info("Job created %s", job_uid)

    root_span = pipeline_run.get_root_span()

    return root_span.create_child_span(
        uid=span_uid or job_uid,
        context_data={"time": str(datetime.now())},
        associatedJobUids=[job_uid],
        with_explicit_time=True,
    )


# =====================================================
# Job executions
# =====================================================
@pytest.mark.integration
@pytest.mark.order(2)
def test_write_file_func(adoc_client):
    span = create_job_span_not_bounded(
        adoc_client,
        "customers.data-generation",
        [],
        [Node(job_uid="customers.s3-upload")],
        JobMetadata("Jason", "COKE", "https://github.com/coke/reports/customers.kt"),
        {"job": "data_gene"},
        "customers.data.generation",
    )

    span.start(created_at=child_span_1_created_at)

    span.send_event(
        GenericEvent(
            context_data={
                "Size": test_const.CSV_SIZE - 1,
                "total_file": 1,
            },
            event_uid="customers.data.generation.metadata",
            created_at=child_span_1_created_at,
        )
    )

    span.send_event(
        LogEvent(
            context_data={"Size": test_const.CSV_SIZE - 1},
            log_data="Customer data generated successfully",
            created_at=child_span_1_created_at,
        )
    )

    span.end(created_at=child_span_1_finished_at)
    logger.info("Data generation completed")


@pytest.mark.integration
@pytest.mark.order(3)
def test_upload_file_to_s3(adoc_client):
    span = create_job_span_not_bounded(
        adoc_client,
        "customers.s3-upload",
        [Node(job_uid="customers.data-generation")],
        [Node(asset_uid=f"{test_const.S3_DS}.{test_const.S3_CUSTOMER}")],
        JobMetadata("Jason", "COKE", "https://github.com/coke/reports/customers.kt"),
        {"job": "data_upload"},
        "customers.s3.upload",
    )

    span.start(created_at=child_span_2_created_at)
    span.end(created_at=child_span_2_finished_at)
    logger.info("Upload to S3 completed")

@pytest.mark.integration
@pytest.mark.order(4)
def test_rds_and_s3_clubbing(adoc_client):
    span = create_job_span_not_bounded(
        adoc_client,
        "customers.s3-snowflake-clubbing",
        [
            Node(asset_uid=f"{test_const.S3_DS}.{test_const.S3_CUSTOMER}"),
            Node(asset_uid=f"{test_const.SDK_SNOWFLAKE_DATA_SOURCE}.{test_const.SNOWFLAKE_SERVICES}"),
        ],
        [Node(asset_uid=f"{test_const.SDK_SNOWFLAKE_DATA_SOURCE}.{test_const.S3_SNOWFLAKE_CUSTOMERS}")],
        JobMetadata("BEN", "COKE", "https://github.com/coke/reports/rds_customers.kt"),
        {"job": "data_clubbing"},
        "customers.s3.snowflake.clubbing",
    )

    span.start(created_at=child_span_3_created_at)
    span.end(created_at=child_span_3_finished_at)

    logger.info("RDS + S3 clubbing completed")

@pytest.mark.integration
@pytest.mark.order(5)
def test_close_pipeline_run(adoc_client):
    pipeline = adoc_client.get_pipeline(test_const.EXPLICIT_PIPELINE_UID)
    pipeline_run = pipeline.get_run(pipeline_run_id)

    pipeline_run.get_root_span().end(
        context_data={"dag_status": "SUCCESS"},
        created_at=explicit_pipeline_run_finishedAt,
    )

    pipeline_run.update_pipeline_run(
        context_data={"status": "success"},
        result=PipelineRunResult.SUCCESS,
        status=PipelineRunStatus.COMPLETED,
    )

    logger.info("Pipeline run closed successfully")
